# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Timeline duration control."""
from datetime import datetime
from typing import Iterable, List, Optional, Tuple, Union

import attr
import pandas as pd
from bokeh.io import output_notebook, show
from bokeh.models import ColumnDataSource, HoverTool, LayoutDOM

# pylint: enable=no-name-in-module
from bokeh.plotting import figure, reset_output
from bokeh.layouts import column
from bokeh.transform import dodge

from ..common.utility import check_kwargs
from ..data.query_defns import ensure_df_datetimes
from .timeline import (
    _create_tool_tips,
    _create_range_tool,
    _calc_auto_plot_height,
    _get_time_bounds,
    _set_axes_and_grids,
    _get_tick_formatter,
    _plot_ref_events,
    check_df_columns,
)

# pylint: disable=unused-import
# Importing to activate pandas accessors
from .timeline_pd_accessor import TimeLineAccessor  # noqa F401

# pylint: enable=unused-import

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

_TIMELINE_HELP = (
    "https://msticpy.readthedocs.io/en/latest/msticpy.nbtools.html"
    "#msticpy.nbtools.timeline_duration.{plot_type}"
)


@attr.s(auto_attribs=True)
class PlotParams:
    """Plot params for time_duration."""

    height: Optional[int] = None
    width: int = 900
    title: Optional[str] = None
    yaxis: bool = True
    range_tool: bool = True
    xgrid: bool = True
    ygrid: bool = False
    hide: bool = False
    color: str = "navy"
    ylabel_cols: Iterable[str] = attr.Factory(list)
    ref_events: Optional[pd.DataFrame] = None
    ref_col: Optional[str] = None
    ref_times: Optional[List[Tuple[datetime, str]]] = None
    source_columns: List = []

    @classmethod
    def field_list(cls) -> List[str]:
        """Return field names as a list."""
        return list(attr.fields_dict(cls).keys())


# pylint: disable=too-many-locals


def display_timeline_duration(
    data: pd.DataFrame,
    group_by: Union[Iterable[str], str],
    time_column: str = "TimeGenerated",
    end_time_column: Optional[str] = None,
    **kwargs,
) -> LayoutDOM:  # noqa: C901, MC0001
    """
    Display a duration timeline of events grouped by one or more columns.

    Parameters
    ----------
    data : pd.DataFrame
        Data to plot
    group_by : Union[Iterable[str], str]
        The column name or iterable of column names to group the data by.
    time_column : str
        Primary time column - will be used to calculate the
        start time of the duration for each group.
        If `end_time_column` is not specified it will also be used to
        calculate the end time.
    end_time_column : Optional[str]
        If supplied, it will be used to calculate the end time
        of the duration for each group.

    Other Parameters
    ----------------
    title : str, optional
        Title to display (the default is None)
    ylabel_cols : Optional[Iterable[str]], optional
        The subset of the group columns to use for the y-axis labels.
    yaxis : bool, optional
        Whether to show the yaxis and labels
    range_tool : bool, optional
        Show the the range slider tool (default is True)
    source_columns : list, optional
        List of default source columns to use in tooltips
        (the default is None)
    height : int, optional
        The height of the plot figure
        (the default is auto-calculated height)
    width : int, optional
        The width of the plot figure (the default is 900)
    color : str
        Default series color (default is "navy")
    ref_events : pd.DataFrame, optional
        Add references line/label using the event times in the dataframe.
        (the default is None)
    ref_col : str, optional
        The column name to use for the label from `ref_events`
        (the default is None)
    ref_times : List[Tuple[datetime, str]], optional
        Add one or more reference line/label using (the default is None)

    Returns
    -------
    LayoutDOM
        The bokeh plot figure.

    """
    reset_output()
    output_notebook()

    check_kwargs(kwargs, PlotParams.field_list())
    param = PlotParams(**kwargs)

    group_by = [group_by] if isinstance(group_by, str) else list(group_by)
    end_time_column = end_time_column or time_column
    data = ensure_df_datetimes(data, columns=list(set([time_column, end_time_column])))
    check_df_columns(
        data,
        group_by + [end_time_column, time_column],
        _TIMELINE_HELP,
        "display_timeline_duration",
    )
    grouped_data = _group_durations(data, group_by, time_column, end_time_column)
    min_time = grouped_data["start_time"].min()
    max_time = grouped_data["end_time"].max()

    # Create Bokeh Column Data Sources
    col_src = ColumnDataSource(grouped_data)
    # Re-join original data to grouped data
    all_data = data.merge(grouped_data, on=group_by)
    all_data_src = ColumnDataSource(all_data)

    tool_tip_cols = [*group_by, "start_time", "end_time"]
    if "source_columns" in kwargs:
        tool_tip_cols += kwargs["source_columns"]

    tooltips, formatters = _create_tool_tips(grouped_data, tool_tip_cols)
    hover = HoverTool(tooltips=tooltips, formatters=formatters)

    title = (
        f"Timeline: {param.title}"
        if param.title
        else f"Event Duration Timeline for {', '.join(group_by)}"
    )

    start_range, end_range, min_time, max_time = _get_time_bounds(min_time, max_time)
    height = param.height or _calc_auto_plot_height(len(grouped_data))
    # Concatenate ylabel columns to display on y-axis
    if len(group_by) > 1:
        y_range = grouped_data[group_by[0]].str.cat(
            grouped_data[group_by[1:]], sep=" / "
        )
    else:
        y_range = grouped_data[group_by[0]]

    plot = figure(
        x_range=(start_range, end_range),
        y_range=y_range,
        min_border_left=50,
        plot_height=height,
        plot_width=param.width,
        x_axis_label="Event Time",
        y_axis_label=", ".join(group_by),
        x_axis_type="datetime",
        x_minor_ticks=10,
        tools=[hover, "xwheel_zoom", "box_zoom", "reset", "save", "xpan"],
        title=title,
    )

    # Plot the duration rectangles
    rect_plot_params = dict(
        height=0.3, source=col_src, fill_alpha=0.4, color=param.color
    )
    plot.rect(x="Center", y=dodge("Row", 0.5), width="Width", **rect_plot_params)

    # Plot the individual events as diamonds
    plot.diamond(
        x=time_column,
        y=dodge("Row", 0.5),
        color=param.color,
        alpha=0.5,
        size=5,
        source=all_data_src,
    )

    # Set grid parameters
    _set_axes_and_grids(None, plot, param.yaxis, param.ygrid, param.xgrid)

    # Create plot bar to act as as range selector
    rng_select = _create_range_tool(
        data=all_data,
        min_time=min_time,
        max_time=max_time,
        plot_range=plot.x_range,
        width=param.width,
        height=height,
        time_column=time_column,
        y="Row",
    )

    # set the tick datetime formatter
    plot.xaxis[0].formatter = _get_tick_formatter()
    _plot_ref_events(
        plot=plot,
        ref_events=param.ref_events,
        time_col=time_column,
        group_count=len(grouped_data),
        ref_col=param.ref_col,
        ref_times=param.ref_times,
    )

    plot_layout = column(plot, rng_select) if param.range_tool else plot
    if not param.hide:
        show(plot_layout)

    return plot_layout


# pylint: disable=too-many-locals


def _group_durations(
    data: pd.DataFrame, group_by: List[str], time_column: str, end_time_column: str
):
    """Group the data and calculate start and end times."""
    grouped_data = data.groupby(group_by).agg(
        start_time=pd.NamedAgg(time_column, "min"),
        end_time=pd.NamedAgg(end_time_column, "max"),
    )
    # If we don't have an endtime for certain types just consider them to be single time events
    for row in grouped_data.iterrows():
        if pd.isnull(row[1]["end_time"]):
            grouped_data.at[row[0], "end_time"] = row[1]["start_time"]
    grouped_data = grouped_data.reset_index()
    grouped_data.index.name = "Row"
    grouped_data = grouped_data.reset_index()
    grouped_data["Width"] = grouped_data["end_time"] - grouped_data["start_time"]
    grouped_data["Center"] = grouped_data["start_time"] + (grouped_data["Width"] / 2)
    return grouped_data



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Pivot pipeline class."""
from collections import namedtuple
from typing import Any, Dict, Iterable, List, Optional

import attr
import pandas as pd
import yaml
from attr import Factory
from tqdm.auto import tqdm

from .._version import VERSION
from . import entities

__version__ = VERSION
__author__ = "Ian Hellen"


_STEP_TYPES = {
    "pivot": "mp_pivot.run",
    "pivot_display": "mp_pivot.display",
    "pivot_tee": "mp_pivot.tee",
    "pivot_tee_exec": "mp_pivot.tee_exec",
    "pd_accessor": None,
}

PipelineExecStep = namedtuple(
    "PipelineExecStep", "accessor, pos_params, params, text, comment"
)


@attr.s(auto_attribs=True)
class PipelineStep:
    """Pivot pipeline step class."""

    name: str
    step_type: str = attr.ib(validator=attr.validators.in_(_STEP_TYPES))
    function: Optional[str] = None
    entity: Optional[str] = None
    comment: Optional[str] = None
    pos_params: List[str] = Factory(list)
    params: Dict[str, Any] = Factory(dict)

    def get_exec_step(self) -> PipelineExecStep:
        """
        Return the executable step details.

        Returns
        -------
        PipelineExecStep
            Named tuple with the following fields
            accessor - the name of the pandas DataFrame accessor function
            params - parameters to be passed to the function
            text - the text representation of the accessor + params
            comment - optional comment that can be used by the pipeline
            builder to add Python comments to output.

        """
        if self.step_type not in _STEP_TYPES:
            raise TypeError(f"Invalid step type {self.step_type}")

        mp_func = _STEP_TYPES[self.step_type]
        accessor = mp_func
        params = self.params
        func_text = f".{mp_func}({self._get_param_string()})"

        if self.step_type == "pivot":
            _, func = _get_entity_and_pivot(self.entity, self.function)
            func_text = (
                f".{mp_func}({self.entity}.{self.function}, {self._get_param_string()})"
            )
            params = {"func": func, **(self.params)}
        elif self.step_type == "pivot_tee_exec":
            func_text = f".{mp_func}('{self.function}', {self._get_param_string()})"
            params = {"df_func": self.function, **(self.params)}
        elif self.step_type == "pd_accessor":
            func_text = f".{self.function}({self._get_param_string()})"
            accessor = self.function

        return PipelineExecStep(
            accessor=accessor,
            pos_params=self.pos_params,
            params=params,
            text=func_text,
            comment=self.comment,
        )

    # pylint: disable=no-member, not-an-iterable
    def _get_param_string(self) -> str:
        """Return text representation of keyword params."""
        pos_params = [
            f"'{param}'" if isinstance(param, str) else str(param)
            for param in self.pos_params
        ]
        params_str = [
            f"{p_name}='{p_val}'"
            for p_name, p_val in self.params.items()
            if isinstance(p_val, str)
        ]
        params_other = [
            f"{p_name}={p_val}"
            for p_name, p_val in self.params.items()
            if not isinstance(p_val, str)
        ]
        return ", ".join(pos_params + params_str + params_other)

    # pylint: enable=no-member, not-an-iterable


def _get_entity_and_pivot(entity_name, func_name):
    """Return the entity and pivot function as objects."""
    entity = getattr(entities, entity_name)
    func_name_path = func_name.split(".")
    obj = entity
    for path in func_name_path:
        obj = getattr(obj, path)
    return entity, obj


def _get_pd_accessor_func(data, df_func):
    """Return the function accessor `df_func` for the DataFrame `df`."""
    acc_name = func_name = func = None
    if "." in df_func:
        acc_name, func_name = df_func.split(".")
        accessor = getattr(data, acc_name, None)
        if accessor:
            func = getattr(accessor, func_name, None)
    else:
        func = getattr(data, df_func, None)
    if func:
        # run the function with any additional args
        return func
    return None


class Pipeline:
    """Pivot pipeline."""

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        steps: Optional[Iterable[PipelineStep]] = None,
    ):
        """
        Create Pipeline instance.

        Parameters
        ----------
        name : str
            The pipeline name.
        description : Optional[str]
            The pipeline description, by default None.
        steps : Optional[Iterable[PipelineStep]]
            Pipeline steps, by default None.

        """
        self.name = name
        self.description = description
        self.steps: List[PipelineStep] = []
        if steps:
            for step in steps:
                self.steps.append(step)

    def __repr__(self) -> str:
        """
        Return string representation of pipeline.

        Returns
        -------
        str
            string representation of pipeline

        """
        step_repr = "\n  ".join(str(step) for step in self.steps)
        return (
            f"Pipeline(name='{self.name}', description='{self.description}', "
            + f"steps=[{step_repr}])"
        )

    @classmethod
    def parse_pipeline(cls, pipeline: Dict[str, Dict[str, Any]]) -> "Pipeline":
        """
        Parse single pipeline from dictionary.

        Parameters
        ----------
        pipeline : Dict[str, Dict[str, Any]]
            Single pipeline as a dictionary:
            {name: {pipeline_dict...}}.

        Returns
        -------
        Pipeline
            The pivot pipeline.

        Raises
        ------
        ValueError
            The dictionary could not be parsed as a pipeline.

        """
        pl_name, pl_dict = next(iter(pipeline.items()))
        if pl_dict and isinstance(pl_dict, dict):
            steps = [PipelineStep(**step) for step in pl_dict.get("steps", [])]
            return cls(
                name=pl_name, description=pl_dict.get("description"), steps=steps
            )
        raise ValueError("Dictionary could not be parsed.")

    @staticmethod
    def parse_pipelines(pipelines: Dict[str, Dict[str, Any]]) -> Iterable["Pipeline"]:
        """
        Parse dict of pipelines.

        Parameters
        ----------
        pipelines : Dict[str, Dict[str, Any]]
            Dict of pipelines.

        Yields
        ------
        Pipeline
            Iterable of pipeline instances

        """
        for p_name, pipeline in pipelines.get("pipelines", {}).items():
            yield Pipeline.parse_pipeline({p_name: pipeline})

    @classmethod
    def from_yaml(cls, yml_str: str) -> Iterable["Pipeline"]:
        """
        Parse pipelines from yaml string.

        Parameters
        ----------
        yml_str : str
            Yaml dict of pipelines.

        Yields
        ------
        Pipeline
            Iterable of pipeline instances

        """
        pipelines = yaml.safe_load(yml_str)
        yield from cls.parse_pipelines(pipelines)

    def to_yaml(self) -> str:
        """
        Return yaml representation of pipeline.

        Returns
        -------
        str
            Pipeline as yaml.

        """
        steps = [attr.asdict(step) for step in self.steps]
        return yaml.dump({self.name: {"description": self.description, "steps": steps}})

    def run(
        self, data: pd.DataFrame, verbose: bool = True, debug: bool = False
    ) -> Optional[Any]:
        """
        Run the pipeline on the supplied DataFrame.

        Parameters
        ----------
        data : pd.DataFrame
            Input DataFrame for pipeline
        verbose : bool, optional
            If True, report progress, by default True
        debug : bool, optional
            If True, report more detailed progress, by default False

        Returns
        -------
        Any
            The output of the last stage of the pipeline

        """
        pipeline_result = data
        pipe_linesteps = tqdm(self.steps, desc="Steps") if verbose else self.steps
        for step in pipe_linesteps:
            exec_action = step.get_exec_step()
            if verbose:
                print("step =", step.name, "\n", exec_action)
            if not isinstance(pipeline_result, pd.DataFrame):
                print(
                    "Output type from previous step is {type(pipeline_result}",
                    "This is not a valid input type for the next stage.",
                )
                break
            exec_kws = {"verbose": verbose, "debug": debug}
            func = _get_pd_accessor_func(pipeline_result, exec_action.accessor)
            pipeline_result = func(
                *exec_action.pos_params, **exec_action.params, **exec_kws
            )

        return pipeline_result

    def print_pipeline(self, df_name: str = "input_df", comments: bool = True) -> str:
        """
        Return the pipeline as text that can be executed in Python.

        Parameters
        ----------
        df_name : str, optional
            Name of the input dataframe to be used in the returned
            code, by default "input_df"
        comments : bool, optional
            If True show step comments, by default True

        Returns
        -------
        str
            The executable pipeline text.

        """
        step_list = []
        if comments:
            step_list.append(f"# {self.description or self.name}")
        step_list.extend(["(", f"    {df_name}"])
        for step in self.steps:
            exec_action = step.get_exec_step()
            if comments:
                step_list.append(f"    # {step.comment or step.name}")
            step_list.append(f"    {exec_action.text}")
        step_list.append(")")
        return "\n".join(step_list)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Pivot browser widget."""
from typing import Dict, List

from IPython import get_ipython
from IPython.display import display
import ipywidgets as widgets
from . import entities

from .._version import VERSION

try:
    import pyperclip

    _ENABLE_CLIP = True
except ImportError:
    _ENABLE_CLIP = False

__version__ = VERSION
__author__ = "Ian Hellen"


def _get_entities_with_pivots():
    piv_entities = {}
    for entity_name in dir(entities):
        entity = getattr(entities, entity_name)
        if not isinstance(entity, type) or not issubclass(entity, entities.Entity):
            continue
        piv_list = sorted(entity.get_pivot_list())
        if piv_list:
            piv_entities[entity_name] = entity, piv_list
    return piv_entities


def _box_layout(width="95%"):
    return widgets.Layout(
        **{
            "width": width,
            "border": "solid gray 1px",
            "margin": "1pt",
            "padding": "5pt",
        }
    )


_GENERIC_HELP = """
<p><b>Call with string/simple value:</b></p>
<p><i>Entity.categ.func</i>(<i>arg</i>)</p>
<p><code>IpAddress.util.whois("192.168.1.1")</code></p>

<p><b>Call with list:</b>
<p><i>Entity.categ.func</i>(<i>list</i>)</p>
<p><code>IpAddress.util.ip_type(ip_list1)</code></p>

<p><b>Call with entity:</b></p>
<p><i>Entity.categ.func</i>(<i>entity</i>)</p>
<p><code>IpAddress.util.ip_type(ip_entity)</code></p>

<p><b>Call with dataframe:</b></p>
<p><i>Entity.categ.func</i>(data=<i>df</i>, column=<i>col_name</i>)</p>
<p><code>IpAddress.util.ip_type(data=ip_df, column="ip")</code></p>
<hr>
<p><b>Note:</b> when calling data queries, instead of using generic parameter
names like 'column' or value, you must use the parameter name
expected by the function. E.g. '..func(host_name="myhost")',
func(data=my_df, account_name="ian")', etc.</p>
<p>See the function-specific help for more details</p>

"""


class PivotBrowser:
    """Pivot enitity browser class."""

    def __init__(self):
        """Create an instance of the Pivot browser."""
        self._text: Dict[str, widgets.Widget] = {}
        self._select: Dict[str, widgets.Widget] = {}
        self._layout: Dict[str, widgets.Widget] = {}
        self._html: Dict[str, widgets.Widget] = {}
        self._btn: Dict[str, widgets.Widget] = {}

        self.piv_entities: Dict[str, List[str]] = _get_entities_with_pivots()

        self._create_select_controls()
        self._create_help_controls()
        self._create_search_controls()

        deflt_funcs = next(iter(self.piv_entities.values()))[1][0]
        self._select_function({"new": deflt_funcs})
        self._search_func(change={"new": ""})
        self.layout = widgets.VBox(
            [
                self._layout["entity_funcs"],
                self._layout["search_box"],
                self._layout["disp_btn_hbox"],
                self._layout["func_help_accd"],
            ]
        )

    def _create_select_controls(self):
        """Instantiate and set up select widgets."""
        self._select["pivot_funcs"] = widgets.Select(
            description="pivot function",
            options=next(iter(self.piv_entities.values()))[1],
            layout=widgets.Layout(height="300px", width="95%"),
        )

        self._select["pivot_funcs"].observe(self._select_function, names="value")

        self._select["entities"] = widgets.Select(
            description="entity",
            options=list(self.piv_entities.keys()),
            layout=widgets.Layout(height="300px", width="70%"),
        )

    def _create_help_controls(self):
        """Instantiate and set up help widgets."""
        self._html["gen_help"] = widgets.HTML(_GENERIC_HELP)
        self._html["func_help"] = widgets.HTML("")
        self._html["search_res"] = widgets.HTML("")
        self._layout["func_help_accd"] = widgets.Accordion(
            children=[
                self._html["gen_help"],
                self._html["func_help"],
                self._html["search_res"],
            ],
            layout=_box_layout("95%"),
        )
        self._layout["func_help_accd"].set_title(0, "Generic Pivot function help")
        self._layout["func_help_accd"].set_title(1, "Function-specific help")
        self._layout["func_help_accd"].set_title(2, "Search results (0)")
        self._layout["func_help_accd"].selected_index = None

        self._select["entities"].observe(self._display_pivots, names="value")

        self._layout["entity_box"] = widgets.VBox(
            [widgets.HTML("<b>Entities</b>"), self._select["entities"]],
            layout=_box_layout("40%"),
        )
        self._layout["funcs_box"] = widgets.VBox(
            [
                widgets.HTML("<b>Selected entity pivot Functions</b>"),
                self._select["pivot_funcs"],
            ],
            layout=_box_layout("55%"),
        )
        self._layout["entity_funcs"] = widgets.HBox(
            [self._layout["entity_box"], self._layout["funcs_box"]],
            layout=widgets.Layout(width="100%"),
        )

    def _create_search_controls(self):
        """Instantiate and set up search widgets."""
        self._text["search_txt"] = widgets.Text(
            description="Search:",
            style={"description_width": "initial"},
            layout=widgets.Layout(width="95%"),
            continuous_update=False,
        )
        self._layout["search_box"] = widgets.HBox(
            [self._text["search_txt"]], layout=_box_layout("95%")
        )
        self._text["search_txt"].observe(self._search_func, names="value")
        self._html["cur_func_title"] = widgets.HTML(
            description="Current function:", style={"description_width": "initial"}
        )
        self._btn["copy"] = widgets.Button(description="Copy to clipboard")
        self._btn["copy"].on_click(self._copy_to_clip)
        self._btn["copy"].enabled = _ENABLE_CLIP
        self._layout["disp_btn_hbox"] = widgets.HBox(
            [self._html["cur_func_title"], self._btn["copy"]], layout=_box_layout("95%")
        )

    def display(self):
        """Display the widget."""
        display(self.layout)

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()

    def _display_pivots(self, change):
        entity = change.get("new")
        self._select["pivot_funcs"].options = self.piv_entities.get(entity)[1]

    def _select_function(self, change):
        func_name = change.get("new")
        entity_name = self._select["entities"].value
        entity = self.piv_entities[entity_name][0]
        func_name_path = func_name.split(".")
        obj = entity
        for path in func_name_path:
            obj = getattr(obj, path)
        doc_str = [
            "<i>Note: use generic help and parameters listed "
            + "there for most common use of pivot functions</i>",
            "",
            f"<b>{entity_name}.{func_name}</b>",
            obj.__doc__.replace("\n", "<br>").replace(" ", "&nbsp;"),
        ]

        self._html["func_help"].value = "<br>".join(doc_str)
        self._html["cur_func_title"].value = f"<code>{self._get_current_func()}</code>"

    def _get_current_func(self):
        func_name = self._select["pivot_funcs"].value
        entity_name = self._select["entities"].value
        return f"entities.{entity_name}.{func_name}()"

    def _add_cur_func_to_cell(self, btn):
        del btn
        get_ipython().set_next_input(self._get_current_func())

    def _copy_to_clip(self, btn):
        del btn
        if _ENABLE_CLIP:
            pyperclip.copy(self._get_current_func())

    def _search_func(self, change):
        search_txt = change.get("new")
        search_results, hits = self._search_funcs(search_txt)
        self._layout["func_help_accd"].set_title(2, f"Search results ({hits})")
        self._html["search_res"].value = search_results

    def _search_funcs(self, search_txt):
        f_entities = {
            ent for ent in self.piv_entities if search_txt.casefold() in ent.casefold()
        }
        matching_funcs = [
            (ent, func)
            for ent, ent_piv in self.piv_entities.items()
            for func in ent_piv[1]
            if search_txt.casefold() in func.casefold()
        ]

        hit_count = len(f_entities)
        ent_with_funcs = {ent[0] for ent in matching_funcs}
        f_entities.update(ent_with_funcs)

        ent_func_list = []
        for entity in f_entities:
            ent_func_list.append(f"<b>{entity}</b>")
            if entity in ent_with_funcs:
                ent_funcs = [item[1] for item in matching_funcs if item[0] == entity]
                ent_func_list.extend(ent_funcs)
                hit_count += len(ent_funcs)
        return "<br>".join(ent_func_list), hit_count



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""OData Driver class."""
import abc
import re
import urllib
from typing import Any, Dict, Iterable, Optional, Tuple, Union

import pandas as pd
import requests

from ..._version import VERSION
from ...common import pkg_config as config
from ...common.provider_settings import get_provider_settings
from ...common.exceptions import MsticpyConnectionError, MsticpyUserConfigError
from .driver_base import DriverBase, QuerySource

__version__ = VERSION
__author__ = "Pete Bryan"

_HELP_URI = (
    "https://msticpy.readthedocs.io/en/latest/data_acquisition"
    "/DataProviders.html#connecting-to-an-odata-source"
)

# pylint: disable=too-many-instance-attributes


class OData(DriverBase):
    """Parent class to retreive date from an oauth based API."""

    CONFIG_NAME = ""
    _ALT_CONFIG_NAMES: Iterable[str] = []

    def __init__(self, **kwargs):
        """
        Instantiate OData driver and optionally connect.

        Parameters
        ----------
        connect: bool, optional
            Set true if you want to connect to the provider at initialization

        """
        super().__init__(**kwargs)
        self.oauth_url: Optional[str] = None
        self.req_body: Optional[Dict[str, Optional[str]]] = None
        self.api_ver: Optional[str] = None
        self.api_root: Optional[str] = None
        self.request_uri: Optional[str] = None
        self.req_headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": None,
        }
        self._loaded = True
        self.aad_token = None
        self._debug = kwargs.get("debug", False)

    @abc.abstractmethod
    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Returns
        -------
        Union[pd.DataFrame, Any]
            A DataFrame (if successfull) or
            the underlying provider result if an error.

        """

    def connect(self, connection_str: str = None, **kwargs):
        """
        Connect to oauth data source.

        Parameters
        ----------
        connection_str: str, optional
            Connect to a data source

        Notes
        -----
        Connection string fields:
        tenant_id
        client_id
        client_secret
        apiRoot
        apiVersion

        """
        cs_dict: Dict[str, Any] = {}
        if connection_str:
            self.current_connection = connection_str
            cs_dict = self._parse_connection_str(connection_str)
        else:
            cs_dict = _get_driver_settings(self.CONFIG_NAME, self._ALT_CONFIG_NAMES)
            # let user override config settings with function kwargs
            cs_dict.update(kwargs)

        missing_settings = [
            setting
            for setting in ("tenant_id", "client_id", "client_secret")
            if setting not in cs_dict
        ]
        if missing_settings:
            raise MsticpyUserConfigError(
                "You must supply the following required connection parameter(s)",
                "to the connect function or add them to your msticpyconfig.yaml.",
                ", ".join(f"'{param}'" for param in missing_settings),
                title="Missing connection parameters.",
                help_uri=("Connecting to OData sources.", _HELP_URI),
            )

        # self.oauth_url and self.req_body are correctly set in concrete
        # instances __init__
        req_url = self.oauth_url.format(tenantId=cs_dict["tenant_id"])  # type: ignore
        req_body = dict(self.req_body)  # type: ignore
        req_body["client_id"] = cs_dict["client_id"]
        req_body["client_secret"] = cs_dict["client_secret"]

        # Authenticate and obtain AAD Token for future calls
        data = urllib.parse.urlencode(req_body).encode("utf-8")
        response = requests.post(url=req_url, data=data)
        json_response = response.json()
        self.aad_token = json_response.get("access_token", None)
        if not self.aad_token:
            raise MsticpyConnectionError(
                f"Could not obtain access token - {json_response['error_description']}"
            )

        self.req_headers["Authorization"] = "Bearer " + self.aad_token
        self.api_root = cs_dict.get("apiRoot", self.api_root)
        if not self.api_root:
            raise ValueError(
                f"Sub class {self.__class__.__name__}", "did not set self.api_root"
            )
        api_ver = cs_dict.get("apiVersion", self.api_ver)
        self.request_uri = self.api_root + str(api_ver)

        print("Connected.")
        self._connected = True

        json_response["access_token"] = None
        return json_response

    # pylint: disable=too-many-branches
    def query_with_results(  # noqa: MC0001
        self, query: str, **kwargs
    ) -> Tuple[pd.DataFrame, Any]:  # noqa: MC0001
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The kql query to execute

        Returns
        -------
        Tuple[pd.DataFrame, results.ResultSet]
            A DataFrame (if successfull) and
            Kql ResultSet.

        """
        if not self.connected:
            self.connect(self.current_connection)
        if not self.connected:
            raise ConnectionError(
                "Source is not connected. ", "Please call connect() and retry."
            )

        if self._debug:
            print(query)

        # Build request based on whether endpoint requires data to be passed in
        # request body in or URL
        if kwargs["body"] is True:
            req_url = self.request_uri + kwargs["api_end"]
            req_url = urllib.parse.quote(req_url, safe="%/:=&?~#+!$,;'@()*[]")
            body = {"Query": query}
            response = requests.post(
                url=req_url, headers=self.req_headers, data=str(body)
            )
        else:
            # self.request_uri set if self.connected
            req_url = self.request_uri + query  # type: ignore
            response = requests.get(url=req_url, headers=self.req_headers)
        if response.status_code != requests.codes["ok"]:
            print(response.json()["error"]["message"])
            if response.status_code == 401:
                raise ConnectionRefusedError(
                    "Authentication failed - possible ", "timeout. Please re-connect."
                )
            # Raise an exception to handle hitting API limits
            if response.status_code == 429:
                raise ConnectionRefusedError("You have likely hit the API limit. ")
            response.raise_for_status()

        json_response = response.json()
        if isinstance(json_response, int):
            print(
                "Warning - query did not complete successfully.",
                "Check returned response.",
            )
            return None, json_response

        if "Results" in json_response:
            result = json_response["Results"]
        else:
            result = json_response

        if not result:
            print("Warning - query did not return any results.")
            return None, json_response
        return pd.json_normalize(result), json_response

    # pylint: enable=too-many-branches

    @staticmethod
    def _parse_connection_str(connection_str: str) -> Dict[str, str]:
        """
        Split connection string components into dictionary.

        Parameters
        ----------
        connection_str : str
            Semi-colon delimited connection string

        Returns
        -------
        Dict[str, str]
            dict of key/pair values

        """
        cs_items = connection_str.split(";")
        return {
            prop[0]: prop[1]
            for prop in [item.strip().split("=") for item in cs_items]
            if prop[0] and prop[1]
        }

    @staticmethod
    def _prepare_param_dict_from_filter(filterstr: str) -> Dict[str, str]:
        """
        Parse filter string into dictionary.

        Parameters
        ----------
        filterstr : str
            OData filter string

        """
        get_params = {}
        for filter_param in re.split(r"[\?\&]+", filterstr):
            if filter_param:
                attr = filter_param.split("=")[0]
                val = filter_param.split("=")[1]
                get_params[attr] = val
        return get_params


_CONFIG_NAME_MAP = {
    "tenant_id": ("tenantid", "tenant_id"),
    "client_id": ("clientid", "client_id"),
    "client_secret": ("clientsecret", "client_secret"),
}


def _map_config_dict_name(config_dict: Dict[str, str]):
    """Map configuration parameter names to expected values."""
    mapped_dict = config_dict.copy()
    for provided_name in config_dict:
        for req_name, alternates in _CONFIG_NAME_MAP.items():
            if provided_name.casefold() in alternates:
                mapped_dict[req_name] = config_dict[provided_name]
                break
    return mapped_dict


def _get_driver_settings(config_name, alt_names) -> Dict[str, str]:
    """Try to retrieve config settings for OAuth drivers."""
    drv_config = get_provider_settings("DataProviders").get(config_name)
    app_config: Dict[str, str] = {}
    if drv_config:
        app_config = dict(drv_config.args)
    else:
        # Otherwise fall back on legacy settings location
        for alt_name in alt_names:
            app_config = config.settings.get(alt_name, {}).get("Args")
            if app_config:
                break

    if not app_config:
        return {}
    # map names to allow for different spellings
    return _map_config_dict_name(app_config)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Helper module for computing training probabilities when modelling sessions."""

from collections import defaultdict
from typing import Tuple, Union, DefaultDict

from ..utils.data_structures import StateMatrix


def compute_cmds_probs(  # nosec
    seq1_counts: Union[StateMatrix, dict],
    seq2_counts: Union[StateMatrix, dict],
    unk_token: str,
) -> Tuple[StateMatrix, StateMatrix]:
    """
    Compute command related probabilities.

    In particular, computes the probabilities for the individual commands,
    and also the probabilities for the transitions of commands.

    Parameters
    ----------
    seq1_counts: Union[StateMatrix, dict]
        individual command counts
    seq2_counts: Union[StateMatrix, dict]
        sequence command (length 2) counts
    unk_token: str
        dummy command to signify an unseen command (e.g. "##UNK##")

    Returns
    -------
    Tuple:
        individual command probabilities,
        sequence command (length 2) probabilities

    """
    total_cmds = sum(seq1_counts.values())

    prior_probs: DefaultDict[str, float] = defaultdict(lambda: 0)
    trans_probs: DefaultDict[str, DefaultDict[str, float]] = defaultdict(
        lambda: defaultdict(lambda: 0)
    )

    # compute prior probs
    for cmd in seq1_counts:
        prior_probs[cmd] = seq1_counts[cmd] / total_cmds

    # compute trans probs
    for prev, currents in seq2_counts.items():
        for current in currents:
            trans_probs[prev][current] = seq2_counts[prev][current] / sum(
                seq2_counts[prev].values()
            )

    prior_probs_sm = StateMatrix(states=prior_probs, unk_token=unk_token)
    trans_probs_sm = StateMatrix(states=trans_probs, unk_token=unk_token)

    return prior_probs_sm, trans_probs_sm


def compute_params_probs(  # nosec
    param_counts: Union[StateMatrix, dict],
    cmd_param_counts: Union[StateMatrix, dict],
    seq1_counts: Union[StateMatrix, dict],
    unk_token: str,
) -> Tuple[StateMatrix, StateMatrix]:
    """
    Compute param related probabilities.

    In particular, computes the probabilities of the individual params,
    and also the probabilities of the params conditional on the command.

    Note that we will be modelling whether a parameter is present or not
    for each command. So we make the modelling assumption that the
    parameters are independent Bernoulii random variables conditional
    on the command.

    Note also that because multiple parameters can appear at a time for
    a command, and because we are computing the probability that each
    parameter is present or not, we do NOT expect the probabilities to
    sum to 1.

    Note also that we use laplace smoothing in the counting
    stage of the calculations. Therefore if you have parameter p which
    appeared for every occurrence of command c, the resulting
    probability for param p appearing conditional on command c would
    NOT equal 1. It would be slightly less due to the laplace smoothing.

    Parameters
    ----------
    param_counts: Union[StateMatrix, dict]
        individual param counts
    cmd_param_counts: Union[StateMatrix, dict]
        param conditional on command counts
    seq1_counts: Union[StateMatrix, dict]
        individual command counts
    unk_token: str
        dummy command to signify an unseen command (e.g. "##UNK##")

    Returns
    -------
    Tuple:
        individual param probabilities,
        param conditional on command probabilities

    """
    param_probs: DefaultDict[str, float] = defaultdict(lambda: 0)
    param_cond_cmd_probs: DefaultDict[str, DefaultDict[str, float]] = defaultdict(
        lambda: defaultdict(lambda: 0)
    )

    for cmd, params in cmd_param_counts.items():
        n_cmd = seq1_counts[cmd]
        for param, count in params.items():
            param_cond_cmd_probs[cmd][param] = count / n_cmd

    tot_cmd = sum(seq1_counts.values())
    for param, count in param_counts.items():
        param_probs[param] = count / tot_cmd

    param_probs_sm = StateMatrix(states=param_probs, unk_token=unk_token)
    param_cond_cmd_probs_sm = StateMatrix(
        states=param_cond_cmd_probs, unk_token=unk_token
    )

    return param_probs_sm, param_cond_cmd_probs_sm


def compute_values_probs(  # nosec
    value_counts: Union[StateMatrix, dict],
    param_value_counts: Union[StateMatrix, dict],
    unk_token: str,
) -> Tuple[StateMatrix, StateMatrix]:
    """
    Compute value related probabilities.

    In particular, compute the probabilities of the individual values,
    and also the probabilities of the values conditional on the param.

    Note that we will be modelling the values as categorical conditional
    on the parameter. Therefore, we DO expect these probabilities to sum
    to 1.

    Note also that each parameter can only take one value at a time
    (unlike how a command can take multiple parameters at a time).

    Parameters
    ----------
    value_counts: Union[StateMatrix, dict]
        individual value counts
    param_value_counts: Union[StateMatrix, dict]
        value conditional on param counts
    unk_token: str
        dummy command to signify an unseen command (e.g. "##UNK##")

    Returns
    -------
    Tuple:
        individual value probabilities,
        value conditional on param probabilities

    """
    value_probs: DefaultDict[str, float] = defaultdict(lambda: 0)
    value_cond_param_probs: DefaultDict[str, DefaultDict[str, float]] = defaultdict(
        lambda: defaultdict(lambda: 0)
    )

    for param, values in param_value_counts.items():
        n_val = sum(values.values())
        for value, count in values.items():
            value_cond_param_probs[param][value] = count / n_val

    tot_val = sum(value_counts.values())
    for value, count in value_counts.items():
        value_probs[value] = count / tot_val

    value_probs_sm = StateMatrix(states=value_probs, unk_token=unk_token)
    value_cond_param_probs_sm = StateMatrix(
        states=value_cond_param_probs, unk_token=unk_token
    )

    return value_probs_sm, value_cond_param_probs_sm



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Process Tree Visualization."""
from typing import Any, Dict, Optional, Union

import pandas as pd

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


def get_process_key(procs: pd.DataFrame, source_index: int) -> str:
    """
    Return the process key of the process given its source_index.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events
    source_index : int, optional
        source_index of the process record

    Returns
    -------
    str
        The process key of the process.

    """
    return procs[procs["source_index"] == source_index].iloc[0].name


# def build_process_key(  # type: ignore  # noqa: F821
#     source_proc: pd.Series,
#     schema: "ProcSchema"
# ) -> str:
#     """
#     Return a process key from a process event.

#     Parameters
#     ----------
#     source_proc : pd.Series, optional
#         Source process
#     schema : ProcSchema, optional
#         The data schema to use, by default None
#         - if None the schema will be inferred

#     Returns
#     -------
#     str
#         Process key of the process

#     """
#     if schema is None:
#         schema = infer_schema(source_proc)
#     proc_path = source_proc[schema.process_name].lower()
#     pid = source_proc[schema.process_id]
#     tstamp = pd.to_datetime(source_proc[schema.time_stamp]).strftime(TS_FMT_STRING)
#     return f"{proc_path}{pid}{tstamp}"


def get_roots(procs: pd.DataFrame) -> pd.DataFrame:
    """
    Return the process tree roots for the current data set.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)

    Returns
    -------
    pd.DataFrame
        Process Tree root processes

    """
    return procs[procs["IsRoot"]]


def get_process(procs: pd.DataFrame, source: Union[str, pd.Series]) -> pd.Series:
    """
    Return the process event as a Series.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row

    Returns
    -------
    pd.Series
        Process row

    Raises
    ------
    ValueError
        If unknown type is supplied as `source`

    """
    if isinstance(source, str):
        return procs.loc[source]
    if isinstance(source, pd.Series):
        return source
    raise ValueError("Unknown type for source parameter.")


def get_parent(
    procs: pd.DataFrame, source: Union[str, pd.Series]
) -> Optional[pd.Series]:
    """
    Return the parent of the source process.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row

    Returns
    -------
    Optional[pd.Series]
        Parent Process row or None if no parent was found.

    """
    proc = get_process(procs, source)
    if proc.parent_key in procs.index:
        return procs.loc[proc.parent_key]
    return None


def get_root(procs: pd.DataFrame, source: Union[str, pd.Series]) -> pd.Series:
    """
    Return the root process for the source process.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row

    Returns
    -------
    pd.Series
        Root process

    """
    proc = get_process(procs, source)
    p_path = proc.path.split("/")
    root_proc = procs[procs["source_index"] == p_path[0]]
    return root_proc.iloc[0]


def get_root_tree(procs: pd.DataFrame, source: Union[str, pd.Series]) -> pd.DataFrame:
    """
    Return the process tree to which the source process belongs.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row

    Returns
    -------
    pd.DataFrame
        Process Tree

    """
    proc = get_process(procs, source)
    p_path = proc.path.split("/")
    return procs[procs["path"].str.startswith(p_path[0])]


def get_tree_depth(procs: pd.DataFrame) -> int:
    """
    Return the depth of the process tree.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)

    Returns
    -------
    int
        Tree depth

    """
    return procs["path"].str.count("/").max() + 1


def get_children(
    procs: pd.DataFrame, source: Union[str, pd.Series], include_source: bool = True
) -> pd.DataFrame:
    """
    Return the child processes for the source process.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row
    include_source : bool, optional
        If True include the source process in the results, by default True

    Returns
    -------
    pd.DataFrame
        Child processes

    """
    proc = get_process(procs, source)
    children = procs[procs["parent_key"] == proc.name]
    if include_source:
        return children.append(proc)
    return children


def get_descendents(
    procs: pd.DataFrame,
    source: Union[str, pd.Series],
    include_source: bool = True,
    max_levels: int = -1,
) -> pd.DataFrame:
    """
    Return the descendents of the source process.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row
    include_source : bool, optional
        Include the source process in the results, by default True
    max_levels : int, optional
        Maximum number of levels to descend, by default -1 (all levels)

    Returns
    -------
    pd.DataFrame
        Descendent processes

    """
    proc = get_process(procs, source)
    descendents = []
    parent_keys = [proc.name]
    level = 0
    rem_procs: Optional[pd.DataFrame] = None
    while max_levels == -1 or level < max_levels:
        if rem_procs is not None:
            # pylint: disable=unsubscriptable-object
            children = rem_procs[rem_procs["parent_key"].isin(parent_keys)]
            rem_procs = rem_procs[~rem_procs["parent_key"].isin(parent_keys)]
            # pylint: enable=unsubscriptable-object
        else:
            children = procs[procs["parent_key"].isin(parent_keys)]
            rem_procs = procs[~procs["parent_key"].isin(parent_keys)]
        if children.empty:
            break
        descendents.append(children)
        parent_keys = children.index
        level += 1

    if descendents:
        desc_procs = pd.concat(descendents)
    else:
        desc_procs = pd.DataFrame(columns=proc.index, index=None)
        desc_procs.index.name = "proc_key"
    if include_source:
        return desc_procs.append(proc).sort_values("path")
    return desc_procs.sort_values("path")


def get_ancestors(procs: pd.DataFrame, source, include_source=True) -> pd.DataFrame:
    """
    Return the ancestor processes of the source process.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row
    include_source : bool, optional
        Include the source process in the results, by default True

    Returns
    -------
    pd.DataFrame
        Ancestor processes

    """
    proc = get_process(procs, source)
    p_path = proc.path.split("/")
    if not include_source:
        p_path.remove(proc.source_index)
    return procs[procs["source_index"].isin(p_path)].sort_values("path")


def get_siblings(
    procs: pd.DataFrame, source: Union[str, pd.Series], include_source: bool = True
) -> pd.DataFrame:
    """
    Return the processes that share the parent of the source process.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)
    source : Union[str, pd.Series]
        source_index of process or the process row
    include_source : bool, optional
        Include the source process in the results, by default True

    Returns
    -------
    pd.DataFrame
        Sibling processes.

    """
    parent = get_parent(procs, source)
    proc = get_process(procs, source)
    siblings = get_children(procs, parent, include_source=False)
    if not include_source:
        return siblings[siblings.index != proc.name]
    return siblings


def get_summary_info(procs: pd.DataFrame) -> Dict[str, int]:
    """
    Return summary information about the process trees.

    Parameters
    ----------
    procs : pd.DataFrame
        Process events (with process tree metadata)

    Returns
    -------
    Dict[str, int]
        Summary statistic about the process tree

    """
    summary: Dict[str, Any] = {}
    summary["Processes"] = len(procs)
    summary["RootProcesses"] = len(procs[procs["IsRoot"]])
    summary["LeafProcesses"] = len(procs[procs["IsLeaf"]])
    summary["BranchProcesses"] = len(procs[procs["IsBranch"]])
    summary["IsolatedProcesses"] = len(procs[(procs["IsRoot"]) & (procs["IsLeaf"])])
    summary["LargestTreeDepth"] = procs["path"].str.count("/").max() + 1
    return summary



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
r"""
eventcluster module.

This module is intended to be used to summarize large numbers of events
into clusters of different patterns. High volume repeating events can
often make it difficult to see unique and interesting items.

The module contains functions to generate clusterable features from
string data. For example, an administration command that does some
maintenance on thousands of servers with a commandline such as:
``install-update -hostname {host.fqdn} -tmp:/tmp/{GUID}/rollback``\  can
be collapsed into a single cluster pattern by ignoring the character
values in the string and using delimiters or tokens to group the values.

This is an unsupervised learning module implemented using SciKit Learn
DBScan.

Contains:
dbcluster_events: generic clustering method using DBSCAN designed to summarize
process events and other similar data by grouping on common features.

add_process_features: derives numerical features from text features such as
commandline and process path.

"""
from binascii import crc32
from functools import lru_cache
from math import log10, floor
import re
from typing import List, Any, Tuple, Union

import numpy as np
import pandas as pd

from ..common.exceptions import MsticpyImportExtraError
from ..common.utility import export
from .._version import VERSION

try:
    from sklearn.cluster import DBSCAN
    from sklearn.preprocessing import Normalizer
    import matplotlib.pyplot as plt
    from matplotlib import cm
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without Sklearn and matplotlib installed",
        title="Error importing Scikit Learn and matplotlib",
        extra="ml",
    ) from imp_err

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-many-arguments, too-many-locals
@export
def dbcluster_events(
    data: Any,
    cluster_columns: List[Any] = None,
    verbose: bool = False,
    normalize: bool = True,
    time_column: str = "TimeCreatedUtc",
    max_cluster_distance: float = 0.01,
    min_cluster_samples: int = 2,
    **kwargs,
) -> Tuple[pd.DataFrame, DBSCAN, np.ndarray]:
    """
    Cluster data set according to cluster_columns features.

    Parameters
    ----------
    data : Any
        Input data as a pandas DataFrame or numpy array
    cluster_columns : List[Any], optional
        List of columns to use for features
        - for DataFrame this is a list of column names
        - for numpy array this is a list of column indexes
    verbose : bool, optional
        Print additional information about clustering results (the default is False)
    normalize : bool, optional
        Normalize the input data (should probably always be True)
    time_column : str, optional
        If there is a time column the output data will be ordered by this
        (the default is 'TimeCreatedUtc')
    max_cluster_distance : float, optional
        DBSCAN eps (max cluster member distance) (the default is 0.01)
    min_cluster_samples : int, optional
        DBSCAN min_samples (the minimum cluster size) (the default is 2)

    Other Parameters
    ----------------
    kwargs: Other arguments are passed to DBSCAN constructor

    Returns
    -------
    Tuple[pd.DataFrame, DBSCAN, np.ndarray]
        Output dataframe with clustered rows
        DBSCAN model
        Normalized data set

    """
    allowed_types = [np.ndarray, pd.DataFrame]

    x_input = None
    if isinstance(data, pd.DataFrame):
        if cluster_columns is None:
            x_input = data.values
        else:
            x_input = data[cluster_columns].values
    elif isinstance(data, np.ndarray):
        x_input = data if cluster_columns is None else data[:, cluster_columns].values
    if x_input is None:
        type_list = ", ".join(str(t) for t in allowed_types)
        raise ValueError(
            f"Input data not in expected format.\n{type(data)}",
            f" is not one of allowed types: {type_list}",
        )

    # Create DBSCAN cluster object
    db_cluster = DBSCAN(
        eps=max_cluster_distance, min_samples=min_cluster_samples, **kwargs
    )

    # Normalize the data (most clustering algorithms don't do well with
    # unnormalized data)
    x_norm = Normalizer().fit_transform(x_input) if normalize else x_input
    # fit the data set
    db_cluster.fit(x_norm)
    labels = db_cluster.labels_
    cluster_set, counts = np.unique(labels, return_counts=True)
    if verbose:
        print(
            "Clustering for set size ",
            len(x_norm),
            " - ",
            len(cluster_set),
            " clusters",
        )
        print("Individual cluster sizes: ", ", ".join(str(c) for c in counts))

    clustered_events = _merge_clustered_items(
        cluster_set, labels, data, time_column, counts
    )

    if verbose:
        print("Cluster output rows: ", len(clustered_events))

    return clustered_events, db_cluster, x_norm


def _merge_clustered_items(
    cluster_set: np.ndarray,
    labels: np.ndarray,
    data: Union[pd.DataFrame, np.ndarray],
    time_column: str,
    counts: np.ndarray,
) -> pd.DataFrame:
    """
    Merge outliers and core clusters into single DataFrame.

    Parameters
    ----------
    cluster_set : np.ndarray
        The set of clusters
    labels : np.ndarray
        The cluster labels
    data : Union[pd.DataFrame, np.ndarray]
        The source data
    time_column : str
        Name of the Time column
    counts : np.ndarray
        The counts of members in each cluster

    Returns
    -------
    pd.DataFrame
        Merged dataframe

    """
    tz_aware = data.iloc[0][time_column].tz if isinstance(data, pd.DataFrame) else False
    ts_type = "datetime64[ns, UTC]" if tz_aware is not None else "datetime64[ns]"

    cluster_list = []
    # Iterate through clusters, adding exemplar to output frame
    # pylint: disable=consider-using-enumerate
    # we need to know the index of the item within the loop
    for idx in range(len(cluster_set)):
        cluster_id = cluster_set[idx]
        class_members = labels == cluster_id
        if isinstance(data, pd.DataFrame):
            time_ordered = data[class_members].sort_values(time_column, ascending=True)
            first_event_time = time_ordered[0:][time_column].iat[0]
            last_event_time = time_ordered[-1:][time_column].iat[0]
        else:
            first_event_time = None
            last_event_time = None

        if cluster_id == -1:
            # 'Noise' events are individual items that could not be assigned
            # to a cluster and so are unique
            cluster_list.append(
                data[class_members]
                .assign(
                    Clustered=False,
                    ClusterId=cluster_id,
                    ClusterSize=1,
                    TimeGenerated=first_event_time,
                    FirstEventTime=first_event_time,
                    LastEventTime=last_event_time,
                )
                .astype(
                    dtype={
                        "TimeGenerated": ts_type,
                        "FirstEventTime": ts_type,
                        "LastEventTime": ts_type,
                    }
                )
            )
        else:
            # Otherwise, just choose the first example of the cluster set
            cluster_list.append(
                data[class_members]
                .assign(
                    Clustered=True,
                    ClusterId=cluster_id,
                    ClusterSize=counts[idx],
                    TimeGenerated=first_event_time,
                    FirstEventTime=first_event_time,
                    LastEventTime=last_event_time,
                )[0:1]
                .astype(
                    dtype={
                        "TimeGenerated": ts_type,
                        "FirstEventTime": ts_type,
                        "LastEventTime": ts_type,
                    }
                )
            )
    # pylint: enable=consider-using-enumerate
    return pd.concat(cluster_list)


@export
def add_process_features(
    input_frame: pd.DataFrame, path_separator: str = None, force: bool = False
) -> pd.DataFrame:
    r"""
    Add numerical features based on patterns of command line and process name.

    Parameters
    ----------
    input_frame : pd.DataFrame
        The input dataframe
    path_separator : str, optional
        Path separator. If not supplied, try to determine
        from 'NewProcessName' column of first 10 rows
        (the default is None)
    force : bool, optional
        Forces re-calculation of feature columns even if they
        already exist (the default is False)

    Returns
    -------
    pd.DataFrame
        Copy of the dataframe with the additional numeric features

    Notes
    -----
    Features added:

    - processNameLen: length of process file name (inc path)
    - processNameTokens: the number of elements in the path
    - processName: the process file name (minus path)
    - commandlineTokens: number of space-separated tokens in the command line
    - commandlineLen: length of the command line
    - commandlineLogLen: log10 length of commandline
    - isSystemSession: 1 if session Id is 0x3e7 for Windows or -1 for Linux
    - commandlineTokensFull: counts number of token separators in commandline
      [\\s\-\\/\.,"\'\|&:;%$()]
    - pathScore: sum of ord() value of characters in path
    - pathLogScore: log10 of pathScore
    - commandlineScore: sum of ord() value of characters in commandline
    - commandlineLogScore: log10 of commandlineScore

    """
    output_df = input_frame.copy()

    # Set any NaN values to empty string
    if "NewProcessName" in output_df and "CommandLine" in output_df:
        output_df[["NewProcessName", "CommandLine"]] = output_df[
            ["NewProcessName", "CommandLine"]
        ].fillna(value="")

    # try to determine the path separator
    if path_separator is None:
        sample_df = output_df.head(10)
        lx_path = len(sample_df[sample_df["NewProcessName"].str.contains("/")])
        path_separator = "/" if lx_path else "\\"
    # Create features from process name and command line
    if "NewProcessName" in output_df:
        _add_processname_features(output_df, force, path_separator)

    if "CommandLine" in output_df:
        _add_commandline_features(output_df, force)

    if "SubjectLogonId" in output_df and ("isSystemSession" not in output_df or force):
        output_df["isSystemSession"] = output_df["SubjectLogonId"].isin(["0x3e7", "-1"])

    return output_df


def _add_processname_features(
    output_df: pd.DataFrame, force: bool, path_separator: str
):
    """
    Add process name default features.

    Parameters
    ----------
    output_df : pd.DataFrame
        The dataframe to add features to
    force : bool
        If True overwrite existing feature columns
    path_separator : str
        Path separator for OS

    """
    if "processName" not in output_df or force:
        output_df["processName"] = output_df.apply(
            lambda x: x.NewProcessName.split(path_separator)[-1], axis=1
        )
    if "pathScore" not in output_df or force:
        output_df["pathScore"] = output_df.apply(
            lambda x: char_ord_score(x.NewProcessName), axis=1
        )
    if "pathLogScore" not in output_df or force:
        output_df["pathLogScore"] = output_df.apply(
            lambda x: log10(x.pathScore) if x.pathScore else 0, axis=1
        )
    if "pathHash" not in output_df or force:
        output_df["pathHash"] = output_df.apply(
            lambda x: crc32_hash(x.NewProcessName), axis=1
        )


def _add_commandline_features(output_df: pd.DataFrame, force: bool):
    """
    Add commandline default features.

    Parameters
    ----------
    output_df : pd.DataFrame
        The dataframe to add features to
    force : bool
        If True overwrite existing feature columns

    """
    if "commandlineLen" not in output_df or force:
        output_df["commandlineLen"] = output_df.apply(
            lambda x: len(x.CommandLine), axis=1
        )
    if "commandlineLogLen" not in output_df or force:
        output_df["commandlineLogLen"] = output_df.apply(
            lambda x: log10(x.commandlineLen) if x.commandlineLen else 0, axis=1
        )
    if "commandlineTokensFull" not in output_df or force:
        output_df["commandlineTokensFull"] = output_df[["CommandLine"]].apply(
            lambda x: delim_count(x.CommandLine), axis=1
        )

    if "commandlineScore" not in output_df or force:
        output_df["commandlineScore"] = output_df.apply(
            lambda x: char_ord_score(x.CommandLine), axis=1
        )
    if "commandlineTokensHash" not in output_df or force:
        output_df["commandlineTokensHash"] = output_df.apply(
            lambda x: delim_hash(x.CommandLine), axis=1
        )


@export
@lru_cache(maxsize=1024)
def delim_count(value: str, delim_list: str = r'[\s\-\\/\.,"\'|&:;%$()]') -> int:
    r"""
    Count the delimiters in input column.

    Parameters
    ----------
    value : str
        Data to process
    delim_list : str, optional
        delimiters to use. (the default is r'[\\s\\\\-\\\\\\\\/\.,"\\\\'|&:;%$()]')

    Returns
    -------
    int
        Count of delimiters in the string.

    """
    return len(re.findall(delim_list, value))


@export
@lru_cache(maxsize=1024)
def delim_hash(value: str, delim_list: str = r'[\s\-\\/\.,"\'|&:;%$()]') -> int:
    r"""
    Return a hash (CRC32) of the delimiters from input column.

    Parameters
    ----------
    value : str
        Data to process
    delim_list : str, optional
        delimiters to use. (the default is r'[\\s\\\\-\\\\\\\\/\.,"\\\\'|&:;%$()]')

    Returns
    -------
    int
        Hash of delimiter set in the string.

    """
    return crc32(bytes("".join(re.findall(delim_list, value)), "utf-8"))


@export
@lru_cache(maxsize=1024)
def char_ord_score(value: str, scale: int = 1) -> int:
    """
    Return sum of ord values of characters in string.

    Parameters
    ----------
    value : str
        Data to process
    scale : int, optional
        reduce the scale of the feature (reducing the
        influence of variations this feature on the clustering
        algorithm (the default is 1)

    Returns
    -------
    int
        [description]

    Notes
    -----
    This function sums the ordinal value of each character in the
    input string. Two strings with minor differences will result in
    a similar score. However, for strings with highly variable content
    (e.g. command lines or http requests containing GUIDs) this may result
    in too much variance to be useful when you are trying to detect
    similar patterns. You can use the scale parameter to reduce the
    influence of features using this function on clustering and anomaly
    algorithms.

    """
    return floor(sum(ord(x) for x in value) / scale)


@export
@lru_cache(maxsize=1024)
def token_count(value: str, delimiter: str = " ") -> int:
    """
    Return count of delimiter-separated tokens pd.Series column.

    Parameters
    ----------
    value : str
        Data to process
    delimiter : str, optional
        Delimiter used to split the column string.
        (the default is ' ')

    Returns
    -------
    int
        count of tokens

    """
    return len(value.split(delimiter))


def _string_score(input_str):
    """Sum the ord(c) for characters in a string."""
    return sum(ord(x) for x in input_str)


@export
@lru_cache(maxsize=1024)
def crc32_hash(value: str) -> int:
    """
    Return the CRC32 hash of the input column.

    Parameters
    ----------
    value : str
        Data to process

    Returns
    -------
    int
        CRC32 hash

    """
    return crc32(bytes(value.encode("utf-8")))


@export
def delim_count_df(
    data: pd.DataFrame, column: str, delim_list: str = r'[\s\-\\/\.,"\'|&:;%$()]'
) -> pd.Series:
    r"""
    Count the delimiters in input column.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to process
    column : str
        The name of the column to process
    delim_list : str, optional
        delimiters to use. (the default is r\'[\\s\\\\-\\\\\\\\/\.,"\\\\'|&:;%$()]\')

    Returns
    -------
    pd.Series
        Count of delimiters in the string in `column`.

    """
    return data[column].str.count(delim_list)


@export
def char_ord_score_df(data: pd.DataFrame, column: str, scale: int = 1) -> pd.Series:
    """
    Return sum of ord values of characters in string.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to process
    column : str
        Column name to process
    scale : int, optional
        reduce the scale of the feature (reducing the
        influence of variations this feature on the clustering
        algorithm (the default is 1)

    Returns
    -------
    pd.Series
        The sum of the ordinal values of the characters
        in `column`.

    Notes
    -----
    This function sums the ordinal value of each character in the
    input string. Two strings with minor differences will result in
    a similar score. However, for strings with highly variable content
    (e.g. command lines or http requests containing GUIDs) this may result
    in too much variance to be useful when you are trying to detect
    similar patterns. You can use the scale parameter to reduce the
    influence of features using this function on clustering and anomaly
    algorithms.

    """
    return data.apply(lambda x: sum(ord(char) for char in x[column]) / scale, axis=1)


@export
def token_count_df(data: pd.DataFrame, column: str, delimiter: str = " ") -> pd.Series:
    """
    Return count of delimiter-separated tokens pd.Series column.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to process
    column : str
        Column name to process
    delimiter : str, optional
        Delimiter used to split the column string.
        (the default is ' ')

    Returns
    -------
    pd.Series
        count of tokens in strings in `column`

    """
    return data.apply(lambda x: len(x[column].split(delimiter)), axis=1)


@export
def crc32_hash_df(data: pd.DataFrame, column: str) -> pd.Series:
    """
    Return the CRC32 hash of the input column.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame to process
    column : str
        Column name to process

    Returns
    -------
    pd.Series
        CRC32 hash of input column

    """
    return data.apply(lambda x: crc32(bytes(x[column].encode("utf-8"))), axis=1)


# pylint: disable=too-many-arguments, too-many-statements
@export  # noqa: C901, MC0001
def plot_cluster(  # noqa: C901, MC0001
    db_cluster: DBSCAN,
    data: pd.DataFrame,
    x_predict: np.ndarray,
    plot_label: str = None,
    plot_features: Tuple[int, int] = (0, 1),
    verbose: bool = False,
    cut_off: int = 3,
    xlabel: str = None,
    ylabel: str = None,
):
    """
    Plot clustered data as scatter chart.

    Parameters
    ----------
    db_cluster : DBSCAN
        DBScan Cluster (from SkLearn DBSCAN).
    data : pd.DataFrame
        Dataframe containing original data.
    x_predict : np.ndarray
        The DBSCAN predict numpy array
    plot_label : str, optional
         If set the column to use to label data points
         (the default is None)
    plot_features :  Tuple[int, int], optional
        Which two features in x_predict to plot (the default is (0, 1))
    verbose : bool, optional
        Verbose execution with some extra info
        (the default is False)
    cut_off : int, optional
        The cluster size below which items are considered outliers
        (the default is 3)
    xlabel : str, optional
        x-axis label (the default is None)
    ylabel : str, optional
        y-axis label (the default is None)

    """
    max_idx = x_predict.shape[1] - 1
    if plot_features[0] >= x_predict.shape[1]:
        raise ValueError(f"plot_features[0] index must be a value from 0 to {max_idx}.")
    if plot_features[1] >= x_predict.shape[1]:
        raise ValueError(f"plot_features[1] index must be a value from 0 to {max_idx}.")
    if plot_features[0] == plot_features[1]:
        mssg = "plot_features indexes must be 2 different values in range 0 to"
        raise ValueError(mssg + f" {max_idx}.")

    labels = db_cluster.labels_
    core_samples_mask = np.zeros_like(labels, dtype=bool)

    # pylint: disable=unsupported-assignment-operation
    # (assignment of numpy array is valid)
    core_samples_mask[db_cluster.core_sample_indices_] = True
    unique_labels = set(labels)

    # pylint: disable=no-member
    # Spectral color map does exist
    colors = [cm.Spectral(each) for each in np.linspace(0, 1, len(unique_labels))]
    # Number of clusters in labels, ignoring noise if present.
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
    n_noise_ = list(labels).count(-1)
    _, counts = np.unique(labels, return_counts=True)

    if verbose:
        print(f"Estimated number of clusters: {n_clusters_}")
        print(f"Estimated number of noise points: {n_noise_}")
        # print("Silhouette Coefficient: %0.3f"
        #       % metrics.silhouette_score(x_predict, labels))

    if (
        not isinstance(data, pd.DataFrame)
        or plot_label is not None
        and plot_label not in data
    ):
        plot_label = None
    p_label = None
    for cluster_id, color in zip(unique_labels, colors):
        if cluster_id == -1:
            # Black used for noise.
            color = [0, 0, 0, 1]
        class_member_mask = labels == cluster_id

        cluster_size = counts[cluster_id]
        marker_size = cluster_size
        marker = "o"
        font_size = "small"
        alpha = 0.4

        if cluster_size < cut_off:
            marker = "+"
            marker_size = 10
            font_size = "large"
            alpha = 1.0
        xy_pos = x_predict[class_member_mask & core_samples_mask]
        plt.plot(
            xy_pos[:, plot_features[0]],
            xy_pos[:, plot_features[1]],
            marker,
            markerfacecolor=tuple(color),
            markersize=marker_size,
        )

        if plot_label:
            first_row = data[class_member_mask].iloc[0]
            if not first_row.empty and plot_label in first_row:
                p_label = first_row[plot_label]
                try:
                    plt.annotate(
                        p_label,
                        xy=(xy_pos[0, plot_features[0]], xy_pos[0, plot_features[1]]),
                        fontsize=font_size,
                        alpha=alpha,
                    )
                except IndexError:
                    pass

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(f"Estimated number of clusters: {n_clusters_}")
    plt.show()
    return plt



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Auditd cluster function."""
import pandas as pd

from .eventcluster import dbcluster_events, add_process_features


from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


def cluster_auditd_processes(audit_data: pd.DataFrame, app: str = None) -> pd.DataFrame:
    """
    Clusters process data into specific processes.

    Parameters
    ----------
    audit_data : pd.DataFrame
        The Audit data containing process creation events
    app: str, optional
        The name of a specific app you wish to cluster

    Returns
    -------
    pd.DataFrame
        Details of the clustered process

    """
    if app is not None:
        processes = audit_data[audit_data["exe"].str.contains(app, na=False)]
    else:
        processes = audit_data
    processes = processes.rename(
        columns={
            "acct": "SubjectUserName",
            "uid": "SubjectUserSid",
            "user": "SubjectUserName",
            "ses": "SubjectLogonId",
            "pid": "NewProcessId",
            "exe": "NewProcessName",
            "ppid": "ProcessId",
            "cmdline": "CommandLine",
        }
    )
    req_cols = [
        "cwd",
        "SubjectUserName",
        "SubjectUserSid",
        "SubjectUserName",
        "SubjectLogonId",
        "NewProcessId",
        "NewProcessName",
        "ProcessId",
        "CommandLine",
    ]
    for col in req_cols:
        if col not in processes:
            processes[col] = ""

    feature_procs_h1 = add_process_features(input_frame=processes)

    clus_events, _, _ = dbcluster_events(
        data=feature_procs_h1,
        cluster_columns=["pathScore", "SubjectUserSid"],
        time_column="TimeGenerated",
        max_cluster_distance=0.0001,
    )
    (
        clus_events.sort_values("TimeGenerated")[
            [
                "TimeGenerated",
                "LastEventTime",
                "NewProcessName",
                "CommandLine",
                "SubjectLogonId",
                "SubjectUserSid",
                "pathScore",
                "isSystemSession",
                "ProcessId",
                "ClusterSize",
            ]
        ].sort_values("ClusterSize", ascending=True)
    )

    procs = clus_events[
        [
            "TimeGenerated",
            "NewProcessName",
            "CommandLine",
            "NewProcessId",
            "SubjectUserSid",
            "cwd",
            "ClusterSize",
            "ProcessId",
        ]
    ]
    procs = procs.rename(columns={"NewProcessId": "pid", "ProcessId": "ppid"})

    return procs



#  -------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for
#  license information.
#  --------------------------------------------------------------------------
"""Splunk Driver class."""
from datetime import datetime
from typing import Any, Tuple, Union, Dict, Iterable, Optional

import pandas as pd

from .driver_base import DriverBase, QuerySource
from ..._version import VERSION
from ...common.utility import export, check_kwargs
from ...common.exceptions import (
    MsticpyConnectionError,
    MsticpyNotConnectedError,
    MsticpyUserConfigError,
    MsticpyImportExtraError,
)
from ...common.provider_settings import get_provider_settings, ProviderSettings

try:
    import splunklib.client as sp_client
    import splunklib.results as sp_results
    from splunklib.client import AuthenticationError, HTTPError
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without splunk-sdk installed",
        title="Error importing splunk-sdk",
        extra="splunk",
    ) from imp_err

__version__ = VERSION
__author__ = "Ashwin Patil"


SPLUNK_CONNECT_ARGS = {
    "host": "(string) The host name (the default is 'localhost').",
    "port": "(integer) The port number (the default is 8089).",
    "http_scheme": "('https' or 'http') The scheme for accessing the service "
    + "(the default is 'https').",
    "verify": "(Boolean) Enable (True) or disable (False) SSL verrification for "
    + "https connections. (optional, the default is True)",
    "owner": "(string) The owner context of the namespace (optional).",
    "app": "(string) The app context of the namespace (optional).",
    "sharing": "('global', 'system', 'app', or 'user') "
    + "The sharing mode for the namespace (the default is 'user').",
    "token": "(string) The current session token (optional). Session tokens can be"
    + " shared across multiple service instances.",
    "cookie": "(string) A session cookie. When provided, you don’t need to call"
    + " login(). This parameter is only supported for Splunk 6.2+.",
    "autologin": "(boolean) When True, automatically tries to log in again if"
    + " the session terminates.",
    "username": "(string) The Splunk account username, which is used to "
    + "authenticate the Splunk instance.",
    "password": "(string) The password for the Splunk account.",
}


@export
class SplunkDriver(DriverBase):
    """Driver to connect and query from Splunk."""

    _SPLUNK_REQD_ARGS = ["host", "username", "password"]
    _CONNECT_DEFAULTS: Dict[str, Any] = {"port": 8089}
    _TIME_FORMAT = '"%Y-%m-%d %H:%M:%S.%6N"'

    def __init__(self, **kwargs):
        """Instantiate Splunk Driver."""
        super().__init__(**kwargs)
        self.service = None
        self._loaded = True
        self._connected = False
        self._debug = kwargs.get("debug", False)
        self.public_attribs = {
            "client": self.service,
            "saved_searches": self._saved_searches,
            "fired_alerts": self._fired_alerts,
        }
        self.formatters = {"datetime": self._format_datetime, "list": self._format_list}

    def connect(self, connection_str: str = None, **kwargs):
        """
        Connect to Splunk via splunk-sdk.

        Parameters
        ----------
        connection_str : Optional[str], optional
            Connection string with Splunk connection parameters

        Other Parameters
        ----------------
        kwargs :
            Connection parameters can be supplied as keyword parameters.

        Notes
        -----
        Default configuration is read from the DataProviders/Splunk
        section of msticpyconfig.yaml, if available.

        """
        cs_dict = self._get_connect_args(connection_str, **kwargs)

        arg_dict = {
            key: val for key, val in cs_dict.items() if key in SPLUNK_CONNECT_ARGS
        }
        try:
            self.service = sp_client.connect(**arg_dict)
        except AuthenticationError as err:
            raise MsticpyConnectionError(
                f"Authentication error connecting to Splunk: {err}",
                title="Splunk connection",
                help_uri="https://msticpy.readthedocs.io/en/latest/DataProviders.html",
            ) from err
        except HTTPError as err:
            raise MsticpyConnectionError(
                f"Communication error connecting to Splunk: {err}",
                title="Splunk connection",
                help_uri="https://msticpy.readthedocs.io/en/latest/DataProviders.html",
            ) from err
        except Exception as err:
            raise MsticpyConnectionError(
                f"Error connecting to Splunk: {err}",
                title="Splunk connection",
                help_uri="https://msticpy.readthedocs.io/en/latest/DataProviders.html",
            ) from err
        self._connected = True
        print("connected")

    def _get_connect_args(
        self, connection_str: Optional[str], **kwargs
    ) -> Dict[str, Any]:
        """Check and consolidate connection parameters."""
        cs_dict: Dict[str, Any] = self._CONNECT_DEFAULTS
        # Fetch any config settings
        cs_dict.update(self._get_config_settings())
        # If a connection string - parse this and add to config
        if connection_str:
            cs_items = connection_str.split(";")
            cs_dict.update(
                {
                    cs_item.split("=")[0].strip(): cs_item.split("=")[1]
                    for cs_item in cs_items
                }
            )
        elif kwargs:
            # if connection args supplied as kwargs
            cs_dict.update(kwargs)
            check_kwargs(cs_dict, list(SPLUNK_CONNECT_ARGS.keys()))

        cs_dict["port"] = int(cs_dict["port"])
        verify_opt = cs_dict.get("verify")
        if isinstance(verify_opt, str):
            cs_dict["verify"] = "true" in verify_opt.casefold()
        elif isinstance(verify_opt, bool):
            cs_dict["verify"] = verify_opt

        missing_args = set(self._SPLUNK_REQD_ARGS) - cs_dict.keys()
        if missing_args:
            raise MsticpyUserConfigError(
                "One or more connection parameters missing for Splunk connector",
                ", ".join(missing_args),
                f"Required parameters are {', '.join(self._SPLUNK_REQD_ARGS)}",
                "All parameters:",
                *[f"{arg}: {desc}" for arg, desc in SPLUNK_CONNECT_ARGS.items()],
                title="no Splunk connection parameters",
            )
        return cs_dict

    def query(
        self, query: str, query_source: QuerySource = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute splunk query and retrieve results via OneShot search mode.

        Parameters
        ----------
        query : str
            Splunk query to execute via OneShot search mode
        query_source : QuerySource
            The query definition object

        Other Parameters
        ----------------
        kwargs :
            Are passed to Splunk oneshot method
            count=0 by default

        Returns
        -------
        Union[pd.DataFrame, Any]
            Query results in a dataframe.
            or query response if an error.

        """
        del query_source
        if not self._connected:
            raise self._create_not_connected_err()
        # default to unlimited query unless count is specified
        count = kwargs.pop("count", 0)
        query_results = self.service.jobs.oneshot(query, count=count, **kwargs)
        reader = sp_results.ResultsReader(query_results)
        resp_rows = [row for row in reader if isinstance(row, dict)]
        if not resp_rows:
            print("Warning - query did not return any results.")
            return [row for row in reader if isinstance(row, sp_results.Message)]
        return pd.DataFrame(resp_rows)

    def query_with_results(self, query: str, **kwargs) -> Tuple[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            Query to execute against splunk instance.

        Returns
        -------
        Union[pd.DataFrame,Any]
            A DataFrame (if successful) or
            the underlying provider result if an error occurs.

        """

    @property
    def service_queries(self) -> Tuple[Dict[str, str], str]:
        """
        Return dynamic queries available on connection to service.

        Returns
        -------
        Tuple[Dict[str, str], str]
            Dictionary of query_name, query_text.
            Name of container to add queries to.

        """
        if not self.connected:
            raise self._create_not_connected_err()
        if hasattr(self.service, "saved_searches") and self.service.saved_searches:
            queries = {
                search.name.strip().replace(" ", "_"): f"search {search['search']}"
                for search in self.service.saved_searches
            }
            return queries, "SavedSearches"
        return {}, "SavedSearches"

    @property
    def driver_queries(self) -> Iterable[Dict[str, Any]]:
        """
        Return dynamic queries available on connection to service.

        Returns
        -------
        Iterable[Dict[str, Any]]
            List of queries with properties: "name", "query", "container"
            and (optionally) "description"

        Raises
        ------
        MsticpyNotConnectedError
            If called before driver is connected.

        """
        if not self.connected:
            raise self._create_not_connected_err()
        if hasattr(self.service, "saved_searches") and self.service.saved_searches:
            return [
                {
                    "name": search.name.strip().replace(" ", "_"),
                    "query": f"search {search['search']}",
                    "query_paths": "SavedSearches",
                    "description": "",
                }
                for search in self.service.saved_searches
            ]
        return []

    @property
    def _saved_searches(self) -> Union[pd.DataFrame, Any]:
        """
        Return list of saved searches in dataframe.

        Returns
        -------
        pd.DataFrame
            Dataframe with list of saved searches with name and query columns.

        """
        if self.connected:
            return self._get_saved_searches()
        return None

    def _get_saved_searches(self) -> Union[pd.DataFrame, Any]:
        """
        Return list of saved searches in dataframe.

        Returns
        -------
        pd.DataFrame
            Dataframe with list of saved searches with name and query columns.

        """
        if not self.connected:
            raise self._create_not_connected_err()
        savedsearches = self.service.saved_searches

        out_df = pd.DataFrame(columns=["name", "query"])

        namelist = []
        querylist = []
        for savedsearch in savedsearches:
            namelist.append(savedsearch.name.replace(" ", "_"))
            querylist.append(savedsearch["search"])
        out_df["name"] = namelist
        out_df["query"] = querylist

        return out_df

    @property
    def _fired_alerts(self) -> Union[pd.DataFrame, Any]:
        """
        Return list of fired alerts in dataframe.

        Returns
        -------
        pd.DataFrame
            Dataframe with list of fired alerts with alert name and count columns.

        """
        if self.connected:
            return self._get_fired_alerts()
        return None

    def _get_fired_alerts(self) -> Union[pd.DataFrame, Any]:
        """
        Return list of fired alerts in dataframe.

        Returns
        -------
        pd.DataFrame
            Dataframe with list of fired alerts with alert name and count columns.

        """
        if not self.connected:
            raise self._create_not_connected_err()
        firedalerts = self.service.fired_alerts

        out_df = pd.DataFrame(columns=["name", "count"])

        alert_names = []
        alert_counts = []
        for alert in firedalerts:
            alert_names.append(alert.name)
            alert_counts.append(alert.count)
        out_df["name"] = alert_names
        out_df["count"] = alert_counts

        return out_df

    # Parameter Formatting methods
    @staticmethod
    def _format_datetime(date_time: datetime) -> str:
        """Return datetime-formatted string."""
        return f'"{date_time.isoformat(sep=" ")}"'

    @staticmethod
    def _format_list(param_list: Iterable[Any]) -> str:
        """Return formatted list parameter."""
        fmt_list = [f'"{item}"' for item in param_list]
        return ",".join(fmt_list)

    # Read values from configuration
    @staticmethod
    def _get_config_settings() -> Dict[Any, Any]:
        """Get config from msticpyconfig."""
        data_provs = get_provider_settings(config_section="DataProviders")
        splunk_settings: Optional[ProviderSettings] = data_provs.get("Splunk")
        return getattr(splunk_settings, "args", {})

    @staticmethod
    def _create_not_connected_err():
        return MsticpyNotConnectedError(
            "Please run the connect() method before running this method.",
            title="not connected to Splunk.",
            help_uri="https://msticpy.readthedocs.io/en/latest/DataProviders.html",
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for pre-defined widget layouts."""
import asyncio
import json
import os
import random
from abc import ABC
from datetime import datetime, timedelta
from enum import IntEnum
from json import JSONDecodeError
from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple, Union, Iterable
from weakref import WeakValueDictionary

import ipywidgets as widgets
import pandas as pd
from deprecated.sphinx import deprecated
from IPython.display import HTML, display
from ipywidgets import Layout

from .._version import VERSION
from ..common.utility import export, check_kwargs
from ..common.timespan import TimeSpan

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-many-lines, invalid-name
class TimeUnit(IntEnum):
    """Time unit enumeration and value."""

    second = 1
    minute = 60
    hour = 60 * 60
    day = 60 * 60 * 24
    week = 7 * 60 * 60 * 24


# pylint: enable=invalid-name


def _parse_time_unit(unit_str: str) -> TimeUnit:
    """Return the TimeUnit enum matching the input string."""
    if unit_str.casefold().startswith("m"):
        return TimeUnit.minute
    if unit_str.casefold().startswith("h"):
        return TimeUnit.hour
    if unit_str.casefold().startswith("s"):
        return TimeUnit.second
    if unit_str.casefold().startswith("d"):
        return TimeUnit.day
    if unit_str.casefold().startswith("w"):
        return TimeUnit.week
    return TimeUnit.minute


_WIDGET_REG: WeakValueDictionary = WeakValueDictionary()


# pylint: disable=too-few-public-methods
class RegisteredWidget(ABC):
    """
    Register widget in the widget registry.

    Registered widgets will store their values in the register.
    Each widget has an ID that that is derived from one or more of the
    initialization parameters. If an instance of the same widget class is
    created with the same parameters, its previous value will be repopulated
    from the registry.
    This is especially useful in notebooks where people accidentally re-run
    the same cell after entering values.
    """

    def __init__(
        self,
        id_vals: Optional[List[Any]] = None,
        val_attrs: Optional[List[str]] = None,
        nb_params: Optional[Dict[str, str]] = None,
        name_space: Dict[str, Any] = globals(),
        register: bool = True,
        **kwargs,
    ):
        """
        Initialize a registered widget.

        Parameters
        ----------
        id_vals : Optional[List[Any]], optional
            The list of parameter values to use to identify this widget instance,
            by default None
        val_attrs : Optional[List[str]], optional
            The names of the attributes to persist in the registry
            and recall, by default ["value"]
        nb_params : Optional[Dict[str, str]], optional
            A dictionary of attribute names and global variables. If the variable
            exists in the global namespace it will be used to populate the
            corresponding widget attribute. This is only done if the widget
            attribute currently has no value (i.e. restoring a value from
            the registry takes priority over this),
            by default None
        name_space : Dict[str, Any], optional
            Namespace to look for global variables, by default None
        register : bool
            Do not register the widget or retrieve values from previously-
            registered instance.

        """
        del kwargs  # allow to be called with kwargs that are ignored
        # Try to retrieve previous values based on ID of this control
        if register and id_vals:
            id_list = [self.__class__.__name__, *[str(val) for val in id_vals]]
            self._id = hash("".join(id_list))

            if not val_attrs:
                val_attrs = ["value"]
            if self._id in _WIDGET_REG:
                for attr in val_attrs:
                    if hasattr(_WIDGET_REG[self._id], attr):
                        setattr(self, attr, getattr(_WIDGET_REG[self._id], attr))
            # register the current instance as the last instance
            _WIDGET_REG[self._id] = self

        # if there are any notebook params relevant to this control
        if nb_params and name_space:
            for attr, nb_param in nb_params.items():
                # if this doesn't have a value set explicitly or
                # one that was recovered from the widget registry
                # set it from the nb_param value
                if nb_param in name_space and not getattr(self, attr, None):
                    setattr(self, attr, name_space[nb_param])


# pylint: enable=too-few-public-methods


@export
class Lookback:
    """Time lookback slider."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        default: Optional[int] = None,
        label: str = "Select time ({units}) to look back",
        origin_time: datetime = None,
        min_value: Optional[int] = None,
        max_value: Optional[int] = None,
        units: str = "hour",
        auto_display: bool = False,
    ):
        """
        Create an instance of the lookback slider widget.

        Parameters
        ----------
        default : int, optional
            The default 'lookback' time (the default is 4)
        label : str, optional
            The description to display
            (the default is 'Select time ({units}) to look back')
        origin_time : datetime, optional
            The origin time (the default is `datetime.utcnow()`)
        min_value : int, optional
            Minimum value (the default is 1)
        max_value : int, optional
            Maximum value (the default is 240)
        units : str, optional
            Time unit (the default is 'hour')
            Permissable values are 'day', 'hour', 'minute', 'second',
            'week'
            These can all be abbreviated down to initial characters
            ('d', 'm', etc.)
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        # default to now
        self.origin_time = datetime.utcnow() if origin_time is None else origin_time

        self._time_unit = _parse_time_unit(units)
        if "{units}" in label:
            label = label.format(units=self._time_unit.name)
        default = _default_before_after(default, self._time_unit)
        min_value = min_value or 1
        max_value = _default_max_buffer(max_value, default, self._time_unit)

        self._lookback_wgt = widgets.IntSlider(
            value=default,
            min=min_value,
            max=max_value,
            step=1,
            description=label,
            layout=Layout(width="60%", height="50px"),
            style={"description_width": "initial"},
        )

        self.end = self.origin_time
        self._time_unit = _parse_time_unit(units)
        self.start = self.end - timedelta(
            seconds=(self._time_unit.value * self._lookback_wgt.value)
        )

        self._lookback_wgt.observe(self._time_range_change, names="value")

        if auto_display:
            self.display()

    @property
    def layout(self):
        """Return underlying widget."""
        return self._lookback_wgt

    def display(self):
        """Display the interactive widgets."""
        display(self._lookback_wgt)

    @property
    def lookback(self):
        """Return current widget lookback value."""
        return self._lookback_wgt.value

    @property
    def value(self):
        """Return current widget lookback value."""
        return self._lookback_wgt.value

    def _time_range_change(self, change):
        del change
        self.start = self.origin_time - timedelta(
            0, self._lookback_wgt.value * self._time_unit.value
        )

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()


def _default_max_buffer(max_default, default, unit) -> int:
    mag_default = abs(int(default * 4))
    if max_default is not None:
        max_value = abs(max_default)
        return max(max_value, mag_default)
    if unit == TimeUnit.day:
        return max(28, mag_default)
    if unit == TimeUnit.hour:
        return max(72, mag_default)
    if unit == TimeUnit.week:
        return max(20, mag_default)
    return max(240, mag_default)


def _default_before_after(default, unit) -> int:
    if default is not None:
        return abs(default)
    if unit in (TimeUnit.day, TimeUnit.week):
        return 1
    if unit == TimeUnit.hour:
        return 6
    return 60


# pylint: disable=too-many-instance-attributes
@export
class QueryTime(RegisteredWidget):
    """
    QueryTime.

    Composite widget to capture date and time origin
    and set start and end times for queries.

    See Also
    --------
    RegisteredWidget

    """

    _ALLOWED_KWARGS = [
        "origin_time",
        "before",
        "after",
        "start",
        "end",
        "max_before",
        "max_after",
        "label",
        "description",
        "units",
        "auto_display",
        "timespan",
        "register",
    ]

    _label_style = {"description_width": "initial"}

    IDS_ATTRIBS = [
        "origin_time",
        "before",
        "after",
        "_query_start",
        "_query_end",
        "_label",
    ]

    def __init__(
        self,
        **kwargs,
    ):
        """
        Create new instance of QueryTime.

        Parameters
        ----------
        origin_time : datetime, optional
            The origin time (the default is `datetime.utcnow()`)
        label : str, optional
            The description to display
            (the default is 'Select time ({units}) to look back')
        before : int, optional
            The default number of `units` before the `origin_time`
            (the default varies based on the unit)
        after : int, optional
            The default number of `units` after the `origin_time`
            (the default varies based on the unit)
        start : Union[datetime, str]
            Start of query time - alternative to specifying origin,
            before, after
        end : Union[datetime, str]
            End of query time - alternative to specifying origin,
            before, after
        timespan : TimeSpan
            TimeSpan of query time - alternative to specifying origin,
            before, after
        max_before : int, optional
            The largest value for `before` (the default varies based on the unit)
        max_after : int, optional
            The largest value for `after` (the default varies based on the unit)
        units : str, optional
            Time unit (the default is 'hour')
            Permissable values are 'day', 'hour', 'minute', 'second',
            'week'
            These can all be abbreviated down to initial characters
            ('d', 'm', etc.)
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        check_kwargs(kwargs, self._ALLOWED_KWARGS)
        self._label = kwargs.pop(
            "description", kwargs.pop("label", "Set query time boundaries")
        )
        self._time_unit = _parse_time_unit(kwargs.get("units", "min"))

        self.before = kwargs.pop("before", None)
        self.after = kwargs.pop("after", None)
        self._query_start = self._query_end = self.origin_time = datetime.utcnow
        self._get_time_parameters(**kwargs)

        self.max_before = kwargs.pop("max_before", None)
        self.max_after = kwargs.pop("max_after", None)
        self._adjust_max_before_after(self.max_before, self.max_after)

        # Call superclass to register
        ids_params = [
            self.origin_time,
            self.before,
            self.after,
            self.max_before,
            self.max_after,
            self._label,
            self._time_unit,
        ]

        super().__init__(id_vals=ids_params, val_attrs=self.IDS_ATTRIBS, **kwargs)

        # Create widgets
        self._w_origin_dt = widgets.DatePicker(
            description="Origin Date", disabled=False, value=self.origin_time.date()
        )
        self._w_origin_tm = widgets.Text(
            description="Time (24hr)",
            disabled=False,
            value=str(self.origin_time.time()),
        )

        range_desc = "Time Range"
        self._w_tm_range = widgets.IntRangeSlider(
            value=(-self.before, self.after),
            min=-self.max_before,
            max=self.max_after,
            step=1,
            description=range_desc,
            disabled=False,
            continuous_update=True,
            orientation="horizontal",
            readout=True,
            readout_format="d",
            layout=Layout(width="70%"),
            style=self._label_style,
        )
        # pylint: disable=no-member
        self._w_time_unit = widgets.Dropdown(
            options=[
                unit for unit, _ in TimeUnit.__members__.items() if unit != "second"
            ],
            value=self._time_unit.name,
            layout=Layout(width="100px"),
        )
        # pylint: enable=no-member

        self._w_start_time_txt = widgets.Text(
            value=self._query_start.isoformat(sep=" "),
            description="Query start time (UTC):",
            layout=Layout(width="50%"),
            style=self._label_style,
        )
        self._w_end_time_txt = widgets.Text(
            value=self._query_end.isoformat(sep=" "),
            description="Query end time (UTC) :  ",
            layout=Layout(width="50%"),
            style=self._label_style,
        )

        # Add change event handlers
        self._w_tm_range.observe(self._time_range_change, names="value")
        self._w_origin_dt.observe(self._update_origin, names="value")
        self._w_origin_tm.observe(self._update_origin, names="value")
        self._w_time_unit.observe(self._change_time_unit, names="value")

        self.layout = self._create_layout()
        if kwargs.pop("auto_display", False):
            self.display()

    def _create_layout(self):
        return widgets.VBox(
            [
                widgets.HTML(f"<h4>{self._label}</h4>"),
                widgets.HBox([self._w_origin_dt, self._w_origin_tm]),
                widgets.VBox(
                    [
                        widgets.HBox([self._w_tm_range, self._w_time_unit]),
                        self._w_start_time_txt,
                        self._w_end_time_txt,
                    ]
                ),
            ]
        )

    def display(self):
        """Display the interactive widgets."""
        display(self.layout)

    def _change_time_unit(self, change):
        """Reset before/after and max buffers to defaults."""
        unit = change["new"]
        self._time_unit = _parse_time_unit(unit)
        self.before = _default_before_after(default=None, unit=self._time_unit)
        self.after = _default_before_after(default=None, unit=self._time_unit)
        self._adjust_max_before_after(max_before=None, max_after=None)
        self._w_tm_range.value = (-self.before, self.after)
        self._w_tm_range.min = -self.max_before
        self._w_tm_range.max = self.max_after

    def _get_time_parameters(self, **kwargs):
        """Process different init time parameters."""
        timespan: TimeSpan = kwargs.pop("timespan", None)
        start = kwargs.pop("start", None)
        end = kwargs.pop("end", None)
        if timespan:
            self._query_end = self.origin_time = timespan.end
            self._query_start = timespan.start
        elif start and end:
            timespan = TimeSpan(start=start, end=end)
            self._query_start = timespan.start
            self._query_end = self.origin_time = timespan.end
        else:
            self.origin_time = kwargs.pop("origin_time", datetime.utcnow())
            self.before = _default_before_after(self.before, self._time_unit)
            self.after = _default_before_after(self.after, self._time_unit)
            # Calculate time offsets from origin
            self._query_start = self.origin_time - timedelta(
                0, self.before * self._time_unit.value
            )
            self._query_end = self.origin_time + timedelta(
                0, self.after * self._time_unit.value
            )
            timespan = TimeSpan(start=self._query_start, end=self._query_end)
        if "units" not in kwargs:
            self._infer_time_units()
        if self.after is None:
            self.after = 0
        if self.before is None:
            self.before = int(
                (self._query_end - self._query_start).total_seconds()
                / self._time_unit.value
            )

    def _infer_time_units(self):
        # If time units not set explicitly, set to something sensible,
        # based on start/end times
        if abs(self.timespan.period.days) > 1:
            self._time_unit = TimeUnit.day
        elif abs(self.timespan.period.total_seconds()) > 3600:
            self._time_unit = TimeUnit.hour
        else:
            self._time_unit = TimeUnit.minute

    def _adjust_max_before_after(self, max_before, max_after):
        """Adjust the max values so the are always bigger than the defaults."""
        self.max_before = _default_max_buffer(
            max_before, self.before or 1, self._time_unit
        )
        self.max_after = _default_max_buffer(
            max_after, self.after or 1, self._time_unit
        )

    def _update_origin(self, change):
        del change
        try:
            tm_value = datetime.strptime(self._w_origin_tm.value, "%H:%M:%S.%f").time()
            self.origin_time = datetime.combine(self._w_origin_dt.value, tm_value)
            self._time_range_change(change=None)
        except (ValueError, TypeError):
            # reset on error
            self._w_origin_dt.value = self.origin_time.date()
            self._w_origin_tm = self.origin_time.time()

    def _time_range_change(self, change):
        del change
        self._query_start = self.origin_time + timedelta(
            0, self._w_tm_range.value[0] * self._time_unit.value
        )
        self._query_end = self.origin_time + timedelta(
            0, self._w_tm_range.value[1] * self._time_unit.value
        )
        self._w_start_time_txt.value = self._query_start.isoformat(sep=" ")
        self._w_end_time_txt.value = self._query_end.isoformat(sep=" ")
        self.before = abs(self._w_tm_range.value[0])
        self.after = abs(self._w_tm_range.value[1])

    @property
    def start(self):
        """Query start time."""
        return self._query_start

    @property
    def end(self):
        """Query end time."""
        return self._query_end

    @property
    def units(self):
        """Time units used by control."""
        return self._time_unit.name

    @property
    def timespan(self):
        """Return the timespan as a TimeSpan object."""
        return TimeSpan(start=self.start, end=self.end)

    @property
    def value(self):
        """Return the timespan as a TimeSpan object."""
        return self.timespan

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()


# pylint: disable=too-many-instance-attributes
@export
class SelectAlert:
    """
    Alert Selector.

    View list of alerts and select one for investigation.
    Optionally provide and action to call with the selected alert as a parameter
    (typically used to display the alert.)

    Attributes
    ----------
    selected_alert : SecurityAlert
        The selected alert
    alert_id : str
        The SystemAlertId of the selected alert
    alerts : List[SecurityAlert]
        The current alert list (DataFrame)
    action : Callable[..., Optional[Tuple[...]]]
        The callback action to execute on selection
        of an alert.

    """

    _ALERTID_REGEX = r"\[id:(?P<alert_id>.*)\]$"

    def __init__(
        self,
        alerts: pd.DataFrame,
        action: Callable[..., Optional[Tuple]] = None,
        columns: List[str] = None,
        time_col: str = "StartTimeUtc",
        id_col: str = "SystemAlertId",
        auto_display: bool = False,
    ):
        """
        Create a new instance of AlertSelector.

        Parameters
        ----------
        alerts : pd.DataFrame
            DataFrame of alerts.
        action : Callable[..., Optional[Tuple]], optional
            Optional function to execute for each selected alert.
            If the function returns one or a tuple of displayable objects
            these will be displayed.
        columns : List[str], optional
            Override the default column names to use from `alerts`
            (the default is ['StartTimeUtc', 'AlertName',
            'CompromisedEntity', 'SystemAlertId'])
        time_col : str, optional
            The column in your alerts that determines when it was created
            Default is 'StartTimeUtc'.
        id_col : str, optional
            The column in your data that determines the alert id
            Default is 'SystemAlertId'.
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        self.alerts = alerts
        self.alert_action = action
        self.id_col = id_col
        self.time_col = time_col

        self.columns = columns or [
            "AlertName",
            "CompromisedEntity",
        ]

        alert_cols = self.columns
        if self.time_col not in alert_cols:
            alert_cols.append(self.time_col)
        if self.id_col not in alert_cols:
            alert_cols.append(self.id_col)
        items = alerts[alert_cols].sort_values(time_col, ascending=True)
        self._select_items = items.apply(
            self._alert_summary,
            axis=1,
            time_col=self.time_col,
            id_col=self.id_col,
            columns=self.columns,
        ).values.tolist()

        self.selected_alert = None
        self.alert_id = None

        self._w_select_alert = widgets.Select(
            options=self._select_items,
            description="Select alert :",
            layout=Layout(width="95%", height="300px"),
            style={"description_width": "initial"},
        )

        self._w_filter_alerts = widgets.Text(
            value="",
            description="Filter alerts by title:",
            style={"description_width": "initial"},
        )

        # setup to use updatable display objects
        rand_id = random.randint(0, 999999)  # nosec
        self._output_id = f"{self.__class__.__name__}_{rand_id}"
        self._disp_elems: List[Any] = []

        # set up observer callbacks
        self._w_filter_alerts.observe(self._update_options, names="value")
        self._w_select_alert.observe(self._select_alert, names="value")
        wgt_list = [self._w_filter_alerts, self._w_select_alert]
        self._w_display_details = widgets.Checkbox(
            value=True,
            description="Display details",
        )
        if action:
            self._w_display_details.observe(self._run_action, names="value")
            wgt_list.append(self._w_display_details)
        self.layout = widgets.VBox(wgt_list)

        if auto_display:
            self.display()

    def display(self):
        """Display the interactive widgets."""
        display(self.layout)
        display(HTML("<hr>"))
        self._select_top_alert()

    @staticmethod
    def _alert_summary(alert_row, time_col, id_col, columns):
        """Return summarized string of alert properties."""
        item = f"{alert_row[time_col]}"
        for col in columns:
            item += f" - {alert_row[col]}"
        item += f" - {alert_row[id_col]}"
        return item

    def _update_options(self, change):
        """Filter the alert list by substring."""
        if change is not None and "new" in change:
            self._w_select_alert.options = [
                alert_dtl
                for alert_dtl in self._select_items
                if change["new"].lower() in alert_dtl[0].lower()
            ]

    def _select_alert(self, selection=None):
        """Select action triggered by picking item from list."""
        if (
            selection is None
            or "new" not in selection
            or not isinstance(selection["new"], str)
        ):
            self.selected_alert = None
        else:
            self.alert_id = selection["new"].split("- ")[-1]

            self.selected_alert = self._get_alert(self.alert_id)
            if self.alert_action is not None:
                self._run_action()

    def _get_alert(self, alert_id):
        """Get the alert by alert_id."""
        self.alert_id = alert_id
        selected_alerts = self.alerts[self.alerts[self.id_col] == alert_id]
        if selected_alerts.shape[0] > 0:
            alert = pd.Series(selected_alerts.iloc[0])
            if "ExtendedProperties" in alert.index and isinstance(
                alert["ExtendedProperties"], str
            ):
                try:
                    alert["ExtendedProperties"] = json.loads(
                        (alert["ExtendedProperties"])
                    )
                except JSONDecodeError:
                    pass
            if "Entities" in alert.index and isinstance(alert["Entities"], str):
                try:
                    alert["Entities"] = json.loads((alert["Entities"]))
                except JSONDecodeError:
                    pass
            return alert
        return None

    def _select_top_alert(self):
        """Select the first alert by default."""
        top_alert = self.alerts.iloc[0]
        if not top_alert.empty:
            self._w_select_alert.value = self._w_select_alert.options[0]
            self.alert_id = top_alert[self.id_col]
            self.selected_alert = self._get_alert(self.alert_id)
            if self.alert_action is not None:
                self._run_action()

    def _run_action(self, change=None):
        """Run any action function and display details, if any."""
        del change
        output_objs = None
        if self._w_display_details.value:
            output_objs = self.alert_action(self.selected_alert)
        if output_objs is None:
            self._clear_display()
            return
        if not isinstance(output_objs, (tuple, list)):
            output_objs = [output_objs]
        display_objs = bool(self._disp_elems)
        for idx, out_obj in enumerate(output_objs):
            if not display_objs:
                self._disp_elems.append(
                    display(out_obj, display_id=f"{self._output_id}_{idx}")
                )
            elif idx == len(self._disp_elems):
                break
            else:
                self._disp_elems[idx].update(out_obj)

    def _clear_display(self):
        """Clear any current details."""
        if not self._disp_elems:
            return
        for disp_obj in self._disp_elems:
            disp_obj.update(HTML(""))

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()


# pylint: disable=too-many-instance-attributes
@deprecated(
    reason="Superceded by SelectAlert. Please use that version", version="0.5.2"
)
@export
class AlertSelector(SelectAlert):
    """
    AlertSelector.

    View list of alerts and select one for investigation.
    Optionally provide and action to call with the selected alert as a parameter
    (typically used to display the alert.)

    Attributes
    ----------
    selected_alert : SecurityAlert
        The selected alert
    alert_id : str
        The SystemAlertId of the selected alert
    alerts : List[SecurityAlert]
        The current alert list (DataFrame)
    action : Callable[..., None]
        The callback action to execute on selection
        of an alert.

    """

    def __init__(
        self,
        alerts: pd.DataFrame,
        action: Callable[..., None] = None,
        columns: List[str] = None,
        auto_display: bool = False,
    ):
        """
        Create a new instance of AlertSelector.

        Parameters
        ----------
        alerts : pd.DataFrame
            DataFrame of alerts.
        action : Callable[..., None], optional
            Optional function to execute for each selected alert.
            (the default is None)
        columns : List[str], optional
            Override the default column names to use from `alerts`
            (the default is ['StartTimeUtc', 'AlertName',
            'CompromisedEntity', 'SystemAlertId'])
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        self._w_output = widgets.Output(layout={"border": "1px solid black"})
        super().__init__(
            alerts=alerts, action=action, columns=columns, auto_display=auto_display
        )

    def display(self):
        """Display the interactive widgets."""
        self._select_top_alert()
        display(
            widgets.VBox([self._w_filter_alerts, self._w_select_alert, self._w_output])
        )

    def _run_action(self, change=None):
        del change
        self._w_output.clear_output()
        with self._w_output:
            self.alert_action(self.selected_alert)

    @property
    def query_params(self):
        """
        Query parameters derived from alert.

        Returns
        -------
            dict(str, str) -- Dictionary of parameter names

        """
        return {"provider_alert_id": self.alert_id}


@export
class GetEnvironmentKey(RegisteredWidget):
    """
    GetEnvironmentKey.

    Tries to retrieve an environment variable value. The value
    can be changed/set and optionally saved back to the system
    environment.
    """

    def __init__(
        self,
        env_var: str,
        help_str: str = None,
        prompt: str = "Enter the value: ",
        auto_display: bool = False,
        **kwargs,
    ):
        """
        Create a new instance of GetEnvironmentKey.

        Parameters
        ----------
        env_var : str
            Name of the environment variable.
        help_str : str, optional
            Help to display if the environment variable is not set. (the default is None)
        prompt : str, optional
            Prompt to display with the text box.
            (the default is "Enter the value: ")
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        env_val = os.environ.get(env_var)
        self._name = env_var
        self._value = ""

        # Call superclass to register
        super().__init__(id_vals=[env_var, prompt], val_attrs=["_value"], **kwargs)

        # Use the registed widget "remembered" value but if the environment
        # variable is set override with this value.
        if env_val is not None:
            self._value = env_val

        if not self._value and help_str is not None:
            display(widgets.HTML(value=help_str))

        self._w_text = widgets.Text(
            value=self._value,
            description=prompt,
            layout=Layout(width="50%"),
            style={"description_width": "initial"},
        )

        self._w_save_button = widgets.Button(
            description="Set",
            layout=Layout(width="10%", disabled=False),
            style={"description_width": "initial"},
        )
        self._w_check_save = widgets.Checkbox(
            value=True, description="Save as environment var", disabled=False
        )
        self._w_save_button.on_click(self._on_save_button_clicked)
        self._hbox = widgets.HBox(
            [self._w_text, self._w_save_button, self._w_check_save]
        )

        if auto_display:
            self.display()

    @property
    def value(self):
        """Get the current value of the key."""
        self._value = self._w_text.value or ""
        return self._value.strip()

    @property
    def name(self):
        """Get the current name of the key."""
        return self._name

    @property
    def layout(self):
        """Return underlying widget collection."""
        return self._hbox

    def display(self):
        """Display the interactive widgets."""
        display(self._hbox)

    def _on_save_button_clicked(self, button):
        del button
        self._value = self.value
        if self._w_check_save.value:
            os.environ[self._name] = self.value

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()


@export
class GetText(RegisteredWidget):
    """
    GetEnvironmentKey.

    Tries to retrieve an environment variable value. The value
    can be changed/set and optionally saved back to the system
    environment.
    """

    def __init__(
        self,
        default: str = None,
        prompt: str = "Enter the value: ",
        auto_display: bool = False,
        **kwargs,
    ):
        """
        Create a new instance of GetEnvironmentKey.

        Parameters
        ----------
        default : str
            Default value.
        prompt : str, optional
            Prompt to display with the text box.
            (the default is "Enter the value: ")
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        See Also
        --------
        RegisteredWidget

        """
        self._value = default

        # Call superclass to register
        super().__init__(id_vals=[default, prompt], val_attrs=["_value"], **kwargs)

        self._w_text = widgets.Text(
            value=self._value,
            description=prompt,
            layout=Layout(width="50%"),
            style={"description_width": "initial"},
        )

        self._w_text.observe(self._update_value, names="value")
        if auto_display:
            self.display()

    def _update_value(self, change):
        self._value = change.get("new", "")

    @property
    def layout(self):
        """Return underlying widget collection."""
        return self._w_text

    @property
    def value(self):
        """Get the current value of the key."""
        return self._value.strip() if self._value else None

    def display(self):
        """Display the interactive widgets."""
        if self._value:
            self._w_text.value = self._value
        display(self.layout)

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()


@export
class SelectItem:
    """
    Selection list from list or dict.

    Attributes
    ----------
    value : Any
        The selected value.
    item_action : Callable[..., Optional[Tuple]]
        Action to call for each selection.

    """

    # pylint: disable=too-many-arguments, too-few-public-methods
    def __init__(
        self,
        description: str = "Select an item",
        item_list: List[str] = None,
        action: Callable[..., Optional[Tuple]] = None,
        item_dict: Mapping[str, str] = None,
        auto_display: bool = False,
        height: str = "100px",
        width: str = "50%",
        display_filter: bool = True,
        value: str = "",
    ):
        """
        Select an item from a list or dict.

        Parameters
        ----------
        description : str, optional
            The widget label to display.
            (the default is 'Select an item')
        item_list : List[str], optional
            A `list` of items to select from (the default is None)
        item_dict : Mapping[str, str], optional
            A `dict` of items to select from. When using `item_dict`
            the keys are displayed as the selectable items and value
            corresponding to the selected key is set as the `value`
            property.
            (the default is None)
        action : Callable[..., Optional[Tuple[...]]], optional
            function to call when item selected (passed a single
            parameter - the value of the currently selected item)
            (the default is None).
            If the function returns one or a tuple of displayable objects
            these will be displayed.
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)
        height : str, optional
            Selection list height (the default is '100px')
        width : str, optional
            Selection list width (the default is '50%')
        display_filter : bool, optional
            Whether to display item filter (the default is True)
        value : str, optional
            A default value to pre-populate the filter with.

        """
        self.def_value = value
        if item_list:
            self._item_list = item_list
            self._item_dict = None
            self.value = item_list[0]
        elif item_dict:
            self._item_list = list(item_dict.keys())
            self._item_dict = item_dict
            self.value = list(self._item_dict.values())[0]
        else:
            raise ValueError("One of item_list or item_dict must be supplied.")

        # Check default value is actually present in our list of items
        list_def_val = (
            self.def_value if self.def_value in self._item_list else self._item_list[0]
        )

        self._wgt_select = widgets.Select(
            value=list_def_val,
            options=self._item_list,
            description=description,
            layout=Layout(width=width, height=height),
            style={"description_width": "initial"},
        )
        self._w_display_details = widgets.Checkbox(
            value=True,
            description="Display details",
        )
        self._display_filter = display_filter
        if display_filter:
            self._w_filter = widgets.Text(
                value=self.def_value,
                description="Filter:",
                style={"description_width": "initial"},
            )

            # set up observer callbacks
            self._w_filter.observe(self._update_options, names="value")
        self._wgt_select.observe(self._select_item, names="value")

        if action:
            self._w_display_details.observe(self._run_action, names="value")
        self.item_action = action

        # setup to use updatable display objects
        rand_id = random.randint(0, 999999)  # nosec
        self._output_id = f"{self.__class__.__name__}_{rand_id}"
        self._disp_elems: List[Any] = []

        if auto_display:
            self.display()

    def _select_item(self, selection):
        if (
            selection is None
            or "new" not in selection
            or not isinstance(selection["new"], str)
        ):
            return
        value = selection["new"]
        self.value = self._item_dict.get(value, None) if self._item_dict else value
        if self.item_action is not None:
            self._run_action()

    def _update_options(self, change):
        """Filter the alert list by substring."""
        if change is not None and "new" in change:
            self._wgt_select.options = [
                i for i in self._item_list if change["new"].lower() in i.lower()
            ]

    def _run_action(self, change=None):
        """Run any action function and display details, if any."""
        del change
        output_objs = None
        if self._w_display_details.value:
            output_objs = self.item_action(self.value)
        if output_objs is None:
            self._clear_display()
            return
        if not isinstance(output_objs, (tuple, list)):
            output_objs = [output_objs]
        display_objs = dict(enumerate(self._disp_elems))
        for idx, out_obj in enumerate(output_objs):
            if idx not in display_objs:
                self._disp_elems.append(
                    display(out_obj, display_id=f"{self._output_id}_{idx}")
                )
            else:
                self._disp_elems[idx].update(out_obj)

    def _clear_display(self):
        """Clear any current details."""
        if not self._disp_elems:
            return
        for disp_obj in self._disp_elems:
            disp_obj.update(HTML(""))

    @property
    def layout(self):
        """Return underlying widget collection."""
        wgt_list = []
        if self._display_filter:
            wgt_list.append(self._w_filter)
        wgt_list.append(self._wgt_select)
        if self.item_action:
            wgt_list.append(self._w_display_details)
        return widgets.VBox(wgt_list)

    def display(self):
        """Display the interactive widget."""
        display(self.layout)
        display(HTML("<hr>"))
        self._show_top_item()

    def _show_top_item(self):
        """Run action on the first item by default."""
        if self.item_action is not None and self.value is not None:
            self._run_action()

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()


@deprecated(reason="Superceded by SelectItem. Please use that version", version="0.5.2")
@export
class SelectString(SelectItem):
    """
    Selection list from list or dict.

    Attributes
    ----------
    value : Any
        The selected value.
    item_action : Callable[..., None]
        Action to call for each selection.

    """

    # pylint: disable=too-many-arguments, too-few-public-methods
    def __init__(
        self,
        description: str = "Select an item",
        item_list: List[str] = None,
        action: Callable[..., None] = None,
        item_dict: Mapping[str, str] = None,
        auto_display: bool = False,
        height: str = "100px",
        width: str = "50%",
        display_filter: bool = True,
    ):
        """
        Select an item from a list or dict.

        Parameters
        ----------
        description : str, optional
            The widget label to display.
            (the default is 'Select an item')
        item_list : List[str], optional
            A `list` of items to select from (the default is None)
        item_dict : Mapping[str, str], optional
            A `dict` of items to select from. When using `item_dict`
            the keys are displayed as the selectable items and value
            corresponding to the selected key is set as the `value`
            property.
            (the default is None)
        action : Callable[..., None], optional
            function to call when item selected (passed a single
            parameter - the value of the currently selected item)
            (the default is None)
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)
        height : str, optional
            Selection list height (the default is '100px')
        width : str, optional
            Selection list width (the default is '50%')
        display_filter : bool, optional
            Whether to display item filter (the default is True)

        """
        self._w_output = widgets.Output(layout={"border": "1px solid black"})

        super().__init__(
            description=description,
            item_list=item_list,
            item_dict=item_dict,
            action=action,
            auto_display=auto_display,
            height=height,
            width=width,
            display_filter=display_filter,
        )

    def _run_action(self, change=None):
        del change
        self._w_output.clear_output()
        with self._w_output:
            self.item_action(self.value)

    def display(self):
        """Display the interactive widget."""
        self._show_top_item()
        wgt_list = []
        if self._display_filter:
            wgt_list.append(self._w_filter)
        wgt_list.append(self._wgt_select)
        if self.item_action:
            wgt_list.append(self._w_output)
        display(widgets.VBox(wgt_list))


@export
class SelectSubset:
    """Class to select a subset from an input list."""

    def __init__(
        self,
        source_items: Union[Dict[str, str], List[Any]],
        default_selected: Union[Dict[str, str], List[Any]] = None,
        display_filter: bool = True,
        auto_display: bool = True,
    ):
        """
        Create instance of SelectSubset widget.

        Parameters
        ----------
        source_items : Union[Dict[str, str], List[Any]]
            List of source items - either a dictionary(label, value),
            a simple list or
            a list of (label, value) tuples.
        default_selected : Union[Dict[str, str], List[Any]]
            Populate the selected list with values - either
            a dictionary(label, value),
            a simple list or
            a list of (label, value) tuples.
        display_filter : bool, optional
            Whether to display item filter (the default is True)
        auto_display : bool, optional
            Whether to display on instantiation (the default is True)

        """
        if isinstance(source_items, dict):
            source_items = list(source_items.items())

        self.src_items = sorted(set(source_items))
        if isinstance(self.src_items[0], tuple):
            self._src_dict = {val: (label, val) for label, val in self.src_items}
        else:
            self._src_dict = {}

        w_layout = widgets.Layout(width="40%", height="200px")
        self._source_list = widgets.SelectMultiple(
            options=sorted(set(self.src_items)), layout=w_layout, description="Source: "
        )

        if isinstance(default_selected, dict):
            default_selected = list(default_selected.items())
        if default_selected:
            set_selected = set(default_selected)
            selected_items = sorted(set_selected.intersection(source_items))
        else:
            selected_items = []

        self._select_list = widgets.SelectMultiple(
            options=selected_items, layout=w_layout, description="Selected: "
        )

        self._display_filter = display_filter
        if display_filter:
            self._w_filter = widgets.Text(
                value="", description="Filter:", style={"description_width": "initial"}
            )

            # set up observer callbacks
            self._w_filter.observe(self._update_options, names="value")

        self._b_add_all = widgets.Button(description="Add All \u21fe")
        self._b_add = widgets.Button(description="Add \u21fe")
        self._b_del = widgets.Button(description="\u21fd Remove")
        self._b_del_all = widgets.Button(description="\u21fd Remove All")

        self._b_add.on_click(self._on_btn_add)
        self._b_del.on_click(self._on_btn_del)
        self._b_del_all.on_click(self._on_btn_del_all)
        self._b_add_all.on_click(self._on_btn_add_all)

        v_box = widgets.VBox(
            [self._b_add_all, self._b_add, self._b_del, self._b_del_all]
        )
        self.layout = widgets.HBox([self._source_list, v_box, self._select_list])
        if self._display_filter:
            self.layout = widgets.VBox([self._w_filter, self.layout])
        if auto_display:
            self.display()

    @property
    def selected_items(self) -> List[Any]:
        """
        Return a list of the selected items.

        If the input list is a list of tuples, this returns
        a list of the selected tuples.

        Returns
        -------
        List[Any]
            List of items in the selected list.

        """
        return list(self._select_list.options)

    @property
    def selected_values(self) -> List[Any]:
        """
        Return list of selected values.

        If the input list is a list of tuples, this returns
        a list of values of the items.

        Returns
        -------
        List[Any]
            List of selected item values.

        """
        if self._select_list.options and isinstance(
            self._select_list.options[0], tuple
        ):
            return [item[1] for item in self._select_list.options]
        return self.selected_items

    def _update_options(self, change):
        """Filter the alert list by substring."""
        if change is not None and "new" in change:
            self._source_list.options = sorted(
                {
                    i
                    for i in self.src_items
                    if str(change["new"]).lower() in str(i).lower()
                }
            )

    # pylint: disable=not-an-iterable
    def _on_btn_add(self, button):
        del button
        selected_set = set(self._select_list.options)
        for selected in self._source_list.value:
            if self._src_dict:
                selected_set.add(self._src_dict[selected])
            else:
                selected_set.add(selected)
        self._select_list.options = sorted(list(selected_set))

    def _on_btn_add_all(self, button):
        del button
        self._select_list.options = sorted(list(set(self._source_list.options)))

    def _on_btn_del(self, button):
        del button
        selected_set = set(self._select_list.options)
        # save the current index
        cur_index = max(self._select_list.index)
        if selected_set:
            for selected in self._select_list.value:
                if self._src_dict:
                    selected_set.remove(self._src_dict[selected])
                else:
                    selected_set.remove(selected)
            self._select_list.options = sorted(list(selected_set))
        if not self._select_list.options:
            return
        # try to set the index to the next item in the list
        if cur_index < len(self._select_list.options):
            next_item = cur_index or 0
            self._select_list.index = tuple([next_item])
        else:
            last_item = max(len(self._select_list.options) - 1, 0)
            self._select_list.index = tuple([last_item])

    # pylint: enable=not-an-iterable

    def _on_btn_del_all(self, button):
        del button
        self._select_list.options = []

    def display(self):
        """Display the control."""
        display(self.layout)

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()


@export
class Progress:
    """UI Progress bar."""

    def __init__(self, completed_len: int, visible: bool = True):
        """
        Instantiate new _Progress UI.

        Parameters
        ----------
        completed_len : int
            The expected value that indicates 100% done.
        visible : bool
            If True start the progress UI visible, by default True.

        """
        self._completed = 0
        self._total = completed_len
        self._progress = widgets.IntProgress(
            value=0,
            max=100,
            step=1,
            description="Progress:",
            bar_style="info",
            orientation="horizontal",
        )
        self._done_label = widgets.Label(value="0%")
        if visible:
            self.show()
        else:
            self.hide()
        self.layout = widgets.HBox([self._progress, self._done_label])
        self.display()

    @property
    def value(self) -> int:
        """
        Return the current progress value.

        Returns
        -------
        int
            Progess value

        """
        return self._completed

    @property
    def max(self) -> int:
        """
        Return the current progress maximum value.

        Returns
        -------
        int
            Max value

        """
        return self._total

    def update_progress(self, new_total: int = 0, delta: int = 0):
        """
        Update progress UI by increment or new total.

        Parameters
        ----------
        new_total : int, optional
            New total, by default 0
        delta : int, optional
            Increment to update current total, by default 0

        """
        if new_total:
            self._completed = new_total
        else:
            self._completed += delta
        perc_total = int(100 * self._completed / self._total)
        self._progress.value = perc_total
        self._done_label.value = f"{perc_total}%"

    def show(self):
        """Make the controls visible."""
        self._hide_show("visible")

    def hide(self):
        """Hide the controls."""
        self._hide_show("hidden")

    def _hide_show(self, visibility):
        vis_layout = widgets.Layout(visibility=visibility)
        self._progress.layout = vis_layout
        self._done_label.layout = vis_layout

    def display(self):
        """Display the control."""
        display(self.layout)

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()


class OptionButtons:
    """
    OptionButtons creates a sequence of buttons to choose from.

    The widget can be run in synchronous mode as a simple option
    selector or in async mode with a timeout.
    In the latter mode, after the timeout has expired the widget
    value is set to the default option button value.
    To use the async mode you must call `widget.display_async()` with
    the async keyword.

    Attributes
    ----------
    value : str
        The value of the option selected (case-normalized)

    Example
    -------
    >>> opt = OptionButtons(description="Continue something?",
    ...  buttons=["Maybe", "Yes", "Cancel"], timeout=10)
    >>> await opt.display_async()

    """

    def __init__(
        self,
        description: Optional[str] = "Select an option to continue",
        buttons: Optional[Iterable[str]] = None,
        default: Optional[str] = None,
        timeout: int = 0,
        debug: bool = False,
    ):
        """
        Initialize the OptionButton widget.

        Parameters
        ----------
        description : Optional[str], optional
            Description label displayed above the buttons,
            by default "Select an option to continue"
        buttons : Optional[Iterable[str]], optional
            A list of button values, by default None. This
            will default to ["Yes", "No", "Cancel"]
        default : Optional[str], optional
            The default value to use on timeout, by default the
            first value in the `buttons` list
        timeout : int, optional
            Timeout in seconds, by default 0
        debug : bool, optional
            Adds some debug information to an Output controle,
            by default False

        """
        if buttons is None:
            buttons = ["Yes", "No", "Cancel"]
        self._buttons = []
        for b_item in buttons:
            self._buttons.append(widgets.Button(description=b_item))
        self._desc_label = widgets.Label(value=description)
        self._timer_label = widgets.Label(layout=widgets.Layout(left="10px"))
        self.default = default or next(iter(buttons)).casefold()
        self.value: Optional[str] = None
        self.timeout = timeout

        self._completion: Any = None
        self._fut_val: Any = None
        self._debug = debug
        if self._debug:
            self._out = widgets.Output()
        self._create_button_callbacks(self._buttons)

    @property
    def layout(self):
        """Create layout for buttons."""
        return widgets.VBox(
            [self._desc_label, widgets.HBox([*(self._buttons), self._timer_label])]
        )

    def _debug_out(self, mssg: str):
        if self._debug:
            self._out.append_stdout(mssg)

    def _create_button_callbacks(self, btns):
        """Set up buttons."""

        def getvalue(change):
            """Button on_click handler."""
            self.value = change.description
            for btn in btns:
                btn.on_click(getvalue, remove=True)

        for btn in btns:
            btn.on_click(getvalue)

    async def _await_widget(self):
        """Awaitable coroutine for widget."""
        self._debug_out("await_widget entered\n")
        self._create_button_callbacks(self._buttons)
        self._debug_out("buttons set\n")

        done, _ = await asyncio.wait(
            [self._wait_for_button_change(), self._await_timer(self.timeout)],
            return_when=asyncio.FIRST_COMPLETED,
            timeout=self.timeout + 5,
        )
        self._debug_out("wait returned\n")
        self._completion = done
        self._debug_out(str(done))
        return done

    async def _wait_for_button_change(self):
        """Awaitable for button selection state."""
        self._debug_out("wait_for_button_change entered\n")
        while self.value is None:
            await asyncio.sleep(0.1)
            if self._debug:
                self._debug_out("*")

    async def _await_timer(self, timeout: int = 5):
        timeout = max(timeout, 0)
        while timeout > 0:
            self._timer_label.value = f"Waiting {timeout} sec..."
            if self.value:
                self._timer_label.value = f"Option selected: '{self.value}'"
                return
            await asyncio.sleep(1)
            timeout -= 1
        self.value = self.default
        self._timer_label.value = f"Timed out. Defaulted to '{self.value}'"

    async def display_async(self, reset: bool = False):
        """
        Display widget with timeout.

        Parameters
        ----------
        reset : bool, optional
            Resets any current value to None,
            by default False

        """
        if reset:
            self.value = None
        display(self.layout)
        if self._debug:
            display(self._out)
        self._fut_val = asyncio.ensure_future(self._await_widget())
        self._debug_out("future returned\n")
        self._debug_out(str(self._fut_val) + "\n")

    def display(self):
        """Display widget in simple sync mode."""
        display(self.layout)

    def _ipython_display_(self):
        """Display in IPython."""
        self.display()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Timespan class."""
from datetime import datetime, timedelta
from typing import Any, Optional, Tuple, Union

import pandas as pd
from dateutil.parser import ParserError  # type: ignore

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


class TimeSpan:
    """Timespan parameter for notebook modules."""

    # pylint: enable=too-many-branches
    def __init__(
        self,
        timespan: Optional[Union["TimeSpan", Tuple[Any, Any], Any]] = None,
        start: Optional[Union[datetime, str]] = None,
        end: Optional[Union[datetime, str]] = None,
        period: Optional[Union[timedelta, str]] = None,
    ):
        """
        Initialize Timespan.

        Parameters
        ----------
        timespan : Union(TimeSpan, Tuple(Any, Any), Any), optional
            A TimeSpan object
            or a tuple of datetimes or datetime strings,
            or an object that has either `start` and `end` or `start` and
            `period` date_time-like attributes.
            By default None
        start : Optional[Union[datetime, str]], optional
            datetime of the start of the time period, by default None
        end : Optional[Union[datetime, str]], optional
            datetime of the end of the time period, by default utcnow
        period : Optional[Union[timedelta, str]], optional
            duration of the period, by default None

        Raises
        ------
        ValueError
            If neither `start` nor `period` are specified.

        """
        start, end, period = self._process_args(timespan, start, end, period)

        if not start and not period:
            raise ValueError(
                "start, period",
                "At least one of 'start' or 'period' must be specified.",
            )

        self._period = None
        if period:
            self._period = self._parse_timedelta(period)

        self._end = self._parse_time(end, "end")
        self._start = self._parse_time(start, "start")
        if self._start and self._period:
            self._end = self._start + self._period
        if self._end is None:
            self._end = datetime.utcnow()
        if self._start is None and self._period:
            self._start = self._end - self._period

    def __eq__(self, value):
        """Return True if the timespans are equal."""
        if not isinstance(value, TimeSpan):
            return False
        return self.start == value.start and self.end == value.end

    def __hash__(self):
        """Return the hash of the timespan."""
        return hash((self.start, self.end))

    def __repr__(self):
        """Return repr string."""
        return (
            f"{self.__class__.__name__}"
            f"(start={self.start}, end={self.end}, period={self.period})"
        )

    @property
    def start(self) -> datetime:
        """
        Return the start of the timeperiod.

        Returns
        -------
        datetime
            Start datetime.

        """
        return self._start

    @property
    def end(self) -> datetime:
        """
        Return the end of the timeperiod.

        Returns
        -------
        datetime
            End datetime.

        """
        return self._end

    @property
    def period(self) -> timedelta:
        """
        Return the period of the timeperiod.

        Returns
        -------
        timedelta
            Period timedelta.

        """
        if not self._period:
            self._period = self.end - self.start
        return self._period

    @staticmethod
    def _process_args(timespan, start, end, period):
        if timespan:
            if isinstance(timespan, TimeSpan):
                start = timespan.start
                end = timespan.end
                period = timespan.period
            elif isinstance(timespan, tuple):
                start = timespan[0]
                end = timespan[1]
        if not start and hasattr(timespan, "start"):
            start = getattr(timespan, "start", None)
        if not end and hasattr(timespan, "end"):
            end = getattr(timespan, "end", None)
        if not period and hasattr(timespan, "period"):
            period = getattr(timespan, "period", None)
        return start, end, period

    @staticmethod
    def _parse_time(time_val, prop_name):
        if time_val is None:
            return None
        if isinstance(time_val, datetime):
            return time_val
        try:
            if isinstance(time_val, str):
                return pd.to_datetime(time_val, infer_datetime_format=True)
        except (ValueError, ParserError):
            pass
        raise ValueError(f"'{prop_name}' must be a datetime or a datetime string.")

    @staticmethod
    def _parse_timedelta(time_val):
        if time_val is None:
            return None
        if isinstance(time_val, timedelta):
            return time_val
        try:
            if isinstance(time_val, str):
                return pd.Timedelta(time_val).to_pytimedelta()
        except (ValueError, ParserError):
            pass
        raise ValueError(
            "'period' must be a pandas-compatible time period string",
            " or Python timedelta.",
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Jupyter Notebook Security Tools."""
# flake8: noqa: F403

# pylint: disable=W0401
from . import nbwidgets

from ..datamodel import entities
from .security_alert import SecurityAlert
from .security_event import SecurityEvent
from .security_alert_graph import *
from ..common import utility as utils

from .observationlist import Observations
from ..common.wsconfig import WorkspaceConfig
from . import nbdisplay

# pylint: enable=W0401

from .._version import VERSION

__version__ = VERSION



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Alert Entity class."""
import json
from datetime import datetime
from typing import Any, Dict, List, Mapping, Optional, Tuple

import pandas as pd

from ..._version import VERSION
from ...common.utility import export
from .entity import Entity, camelcase_property_names

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=invalid-name


@export
class Alert(Entity):
    """
    Alert Entity class.

    Attributes
    ----------
    DisplayName : str
        Alert DisplayName
    CompromisedEntity : str
        Alert CompromisedEntity
    Count : int
        Alert Count
    StartTime : datetime
        Alert StartTime
    EndTime : datetime
        Alert EndTime
    Severity : str
        Alert Severity
    SystemAlertIds : List[str]
        Alert SystemAlertIds
    AlertType : str
        Alert AlertType
    VendorName : str
        Alert VendorName
    ProviderName : str
        Alert ProviderName

    """

    ID_PROPERTIES = ["SystemAlertIds"]

    def __init__(
        self,
        src_entity: Mapping[str, Any] = None,
        src_event: Mapping[str, Any] = None,
        **kwargs,
    ):
        """
        Create a new instance of the entity type.

        Parameters
        ----------
        src_entity : Mapping[str, Any], optional
            Create entity from existing entity or
            other mapping object that implements entity properties.
            (the default is None)
        src_event : Mapping[str, Any], optional
            Create entity from event properties
            (the default is None)


        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Supply the entity properties as a set of
            kw arguments.

        """
        self.DisplayName: Optional[str] = None
        self.CompromisedEntity: Optional[str] = None
        self.Count: Any = None
        self.StartTime: Optional[datetime] = None
        self.EndTime: Optional[datetime] = None
        self.Severity: Any = None
        self.SystemAlertIds: List[str] = []
        self.AlertType: Optional[str] = None
        self.VendorName: Optional[str] = None
        self.ProviderName: Optional[str] = None
        self.Entities: Optional[List] = None
        super().__init__(src_entity=src_entity, **kwargs)
        if src_entity:
            self._create_from_ent(src_entity)

        if isinstance(src_event, pd.Series) or src_event:
            self._create_from_event(src_event)

    def _create_from_ent(self, src_entity):  # noqa: MC0001
        if "StartTime" in src_entity or "TimeGenerated" in src_entity:
            self.TimeGenerated = src_entity["StartTime"] or src_entity["TimeGenerated"]
        if "EndTime" in src_entity:
            self.EndTime = src_entity["EndTime"]
        if "StartTime" in src_entity:
            self.StartTime = src_entity["StartTime"]
        if "AlertDisplayName" in src_entity:
            self.DisplayName = src_entity["AlertDisplayName"]
        if "SystemAlertId" in src_entity:
            self.SystemAlertIds.append(src_entity["SystemAlertId"])
        elif "ID" in src_entity:
            self.SystemAlertIds.append(src_entity["ID"])
        if "Name" in src_entity:
            self.DisplayName = src_entity["Name"]
        if "Entities" in src_entity and src_entity["Entities"]:
            if isinstance(src_entity["Entities"], str):
                try:
                    ents = _extract_entities(json.loads(src_entity["Entities"]))
                except json.JSONDecodeError:
                    ents = []
            else:
                ents = _extract_entities(src_entity["Entities"])
            self.Entities = self._create_entities(ents)
        self._add_additional_data(src_entity)

    def _extract_entities(self, src_row):  # noqa: MC0001
        input_entities = []
        if isinstance(src_row.Entities, str):
            try:
                ext_props = json.loads(src_row["Entities"])
                for item in ext_props:
                    for k, v in item.items():
                        if isinstance(v, dict) and "$ref" in v.keys():
                            item[k] = [x for x in ext_props if x["$id"] == v["$ref"]][0]
                    input_entities.append(item)
            except json.JSONDecodeError:
                pass
        if isinstance(src_row.ExtendedProperties, str):
            try:
                ext_props = json.loads(src_row["ExtendedProperties"])
                for ent, val in ext_props.items():
                    if ent in ["IpAddress", "Username"]:
                        input_entities.append({"Entity": val, "Type": ent})
            except json.JSONDecodeError:
                pass
        return input_entities

    @property
    def description_str(self) -> str:
        """Return Entity Description."""
        if self.StartTime and self.CompromisedEntity:
            return f"{self.DisplayName} ({self.StartTime}) {self.CompromisedEntity}"
        else:
            return f"{self.DisplayName} - {self.SystemAlertIds}"

    @property
    def name_str(self) -> str:
        """Return Entity Name."""
        return f"Alert: {self.DisplayName}" or self.__class__.__name__

    def _add_additional_data(self, src_entity: Mapping[str, Any]):
        """Populate additional alert properties."""
        entity_props = set(self.__dict__.keys()) | {"AlertDisplayName", "SystemAlertId"}
        if isinstance(src_entity, dict):
            prop_list = src_entity.items()
        elif type(src_entity).__name__ == "SecurityAlert":
            prop_list = src_entity.properties.items()  # type: ignore
        # pylint: disable=all
        elif isinstance(src_entity, Mapping):
            prop_list = src_entity.iteritems()  # type: ignore
        # pylint: enable=all
        else:
            return

        for prop_name, prop in prop_list:
            if prop_name not in entity_props:
                self.AdditionalData[prop_name] = prop

    def _create_from_event(self, src_event):
        """Create Alert from an alert event."""
        self.TimeGenerated = src_event.get("StartTime", src_event.get("TimeGenerated"))
        self.DisplayName = src_event.get("DisplayName", src_event.get("Name"))
        self.CompromisedEntity = src_event.get("CompromisedEntity")
        self.StartTime = src_event.get("StartTime")
        self.EndTime = src_event.get("EndTime")
        self.Severity = src_event.get("AlertSeverity")
        self.SystemAlertIds = src_event.get("SystemAlertId", src_event.get("ID"))
        self.AlertType = src_event.get("AlertType")
        self.VendorName = src_event.get("VendorName")
        self.ProviderName = src_event.get("ProviderName")
        if isinstance(src_event["Entities"], str):
            try:
                ents = _extract_entities(json.loads(src_event["Entities"]))
            except json.JSONDecodeError:
                ents = []
        else:
            ents = _extract_entities(src_event["Entities"])
        self.Entities = self._create_entities(ents)

    def _create_entities(self, entities):
        """Create alert entities from returned dicts."""
        new_ents = []
        for ent in entities:
            if isinstance(ent, Tuple):
                ent_details = ent[1]
                ent_type = ent[0]
            elif isinstance(ent, Dict):
                ent_details = ent
                ent_type = ent.get("Type", "Unknown")
            else:
                ent_details = ent
                ent_type = "Unknown"
            new_ent = camelcase_property_names(ent_details)
            ent_obj = Entity.instantiate_entity(
                new_ent, entity_type=Entity.ENTITY_NAME_MAP[ent_type.lower()]
            )
            new_ents.append(ent_obj)
        return new_ents

    _entity_schema = {
        # DisplayName (type System.String)
        "DisplayName": None,
        # CompromisedEntity (type System.String)
        "CompromisedEntity": None,
        # Count (type System.Nullable`1[System.Int32])
        "Count": None,
        # StartTimeUtc (type System.Nullable`1[System.DateTime])
        "StartTime": None,
        # EndTimeUtc (type System.Nullable`1[System.DateTime])
        "EndTime": None,
        # Severity (type System.Nullable`1
        # [Microsoft.Azure.Security.Detection.AlertContracts.V3.Severity])
        "Severity": None,
        # SystemAlertIds (type System.Collections.Generic.List`1[System.String])
        "SystemAlertIds": None,
        # AlertType (type System.String)
        "AlertType": None,
        # VendorName (type System.String)
        "VendorName": None,
        # ProviderName (type System.String)
        "ProviderName": None,
        # List of associated entities
        "Entities": None,
        # Time the alert was generated.
        "TimeGenerated": None,
    }

    def to_html(self, show_entities=False) -> str:
        """Return the item as HTML string."""
        return (
            """
            <h3>Alert: '{name}'</h3>
            <b>Alert_time:</b> {start},
            <b>Compr_entity:</b> {entity},
            <b>Alert_id:</b> {id}
            """.format(
                start=self.properties.get(
                    "StartTimeUtc",
                    self.properties.get("StartTime", "no timestamp"),
                ),
                name=self.properties.get(
                    "AlertDisplayName",
                    self.properties.get("DisplayName", "no alert name"),
                ),
                entity=self.properties.get("CompromisedEntity", "unknown"),
                id=self.properties.get("SystemAlertId", "unknown"),
            )
            if self.properties
            else "Alert has no data."
        )


def _extract_entities(ents: list):
    """Extract all entities from a set and replace $ref elements."""
    base_ents = _generate_base_ents(ents)
    out_ents = []
    for entity in ents:
        if isinstance(entity, dict) and "$ref" in entity:
            out_ents.append(_find_og_ent(entity, base_ents))
        else:
            for k, val in entity.items():
                if isinstance(val, (list, dict)):
                    if isinstance(val, list):
                        nested_ents = []
                        for item in val:
                            if isinstance(item, dict) and "$ref" in item:
                                nested_ents.append(_find_og_ent(item, base_ents))
                                entity[k] = nested_ents
                    elif isinstance(val, dict) and "$ref" in val:
                        entity[k] = _find_og_ent(val, base_ents)
            out_ents.append(entity)
    return out_ents


def _find_og_ent(ent, base_ents):
    """Find the original entity referenced by $ref entity."""
    id = ent["$ref"]
    return next(bent for bent in base_ents if ("$id" in bent) and bent["$id"] == id)


def _generate_base_ents(ents: list) -> list:  # noqa: MC0001
    """Generate a list of all enties form a set of nested entities."""
    base_ents = []
    for ent in ents:
        base_ents.append(ent)
        for _, item in ent.items():
            if isinstance(item, list):
                for prop in item:
                    if isinstance(prop, dict) and "$id" in prop.keys():
                        base_ents.append(prop)
                        for val in prop:
                            if isinstance(prop[val], list):
                                for p in prop[val]:
                                    if isinstance(p, dict) and "$id" in p.keys():
                                        base_ents.append(p)
                            elif (
                                isinstance(prop[val], dict)
                                and "$id" in prop[val].keys()
                            ):
                                base_ents.append(val)
            elif isinstance(item, dict) and "$id" in item.keys():
                base_ents.append(item)
    return base_ents
